###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Publisher root classes.

This includes an abstract `PubRoot` class defining the interface that concrete
classes must implement to support different publisher root sources.  New
classes may be registered with the `register_root_class()` function.

The `make_root()` function, given a target string argument, tries to find the
adequate class that understands the target and can create a publisher root
instance from it.
"""

import io
import pathlib
from abc import ABC, abstractmethod
from collections.abc import AsyncIterator, Callable, Collection, Iterator

try:
    from typing import Self
except ImportError:  # Python < 3.11
    from typing import TypeVar
    Self = TypeVar('Self', bound='PubRoot')

# Requirements
import pydantic


class NoSuchDatasetError(LookupError):
    """The given dataset does not exist."""


class NoSuchChunkError(IndexError):
    """The given chunk does not exist in the dataset."""


class PubRoot(ABC):
    """Abstract class that represents a publisher root."""

    """The class of dataset (relative) paths."""
    Path = pathlib.PurePosixPath

    @classmethod
    @abstractmethod
    def get_maker(cls, target: str) -> Callable[[], Self] | None:
        """Return a callable that returns a root for the given `target`.

        If `target` cannot be used to create an instance of this class,
        return `None`, but no exception should be raised.
        """

    @abstractmethod
    def walk_dsets(self) -> Iterator[Path]:
        """Iterate over the relative paths of datasets in this root."""

    @abstractmethod
    def exists_dset(self, relpath: Path) -> bool:
        """Does the named dataset exist?"""

    @abstractmethod
    def get_dset_etag(self, relpath: Path) -> str:
        """Get a string that varies if the named dataset is modified.

        Raise `NoSuchDatasetError` if the dataset does not exist.
        """

    @abstractmethod
    def get_dset_meta(self, relpath: Path) -> pydantic.BaseModel:
        """Get the metadata of the named dataset.

        Raise `NoSuchDatasetError` if the dataset does not exist.
        """

    @abstractmethod
    def get_dset_chunk(self, relpath: Path, nchunk: int) -> bytes:
        """Get compressed chunk with index `nchunk` of the named dataset.

        Raise `NoSuchDatasetError` if the dataset does not exist.
        Raise `NoSuchChunkError` if the chunk does not exist.
        """

    @abstractmethod
    def open_dset_raw(self, relpath: Path) -> io.BufferedIOBase:
        """Get a byte reader for the raw contents of the named dataset.

        Raise `NoSuchDatasetError` if the dataset does not exist.
        """

    @abstractmethod
    async def awatch_dsets(self) -> AsyncIterator[Collection[Path]]:
        """Yield a set of datasets that have been modified."""


_registered_classes = []


def register_root_class(cls: type) -> bool:
    """Add a publisher root class to the registry.

    This also registers the class as a virtual subclass of `PubRoot`.

    Return whether the class was added or not (because it was already).
    """
    if cls in _registered_classes:
        return False
    PubRoot.register(cls)
    _registered_classes.append(cls)
    return True


class UnsupportedRootError(Exception):
    """No publisher root class supports the given target."""


def make_root(target: str) -> PubRoot:
    """Return a publisher root instance for the given `target`.

    If no registered publisher root class supports the given `target`,
    raise `UnsupportedRootError`.
    """
    for cls in _registered_classes:
        maker = cls.get_maker(target)
        if maker is not None:
            return maker()
    else:
        raise UnsupportedRootError(f"no publisher root class could be used "
                                   f"for target: {target!r}")



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import functools
import io
import logging
import pathlib
import re
from collections.abc import AsyncIterator, Callable, Collection, Iterator, Mapping

try:
    from typing import Self
except ImportError:  # Python < 3.11
    from typing import TypeVar
    Self = TypeVar('Self', bound='PubRoot')    # noqa: F821

# Requirements
import h5py
import pydantic
import watchfiles

# Project
from caterva2 import hdf5
from caterva2.services import pubroot, srv_utils

_MAX_CACHED_CHUNKERS = 32
"""Maximum number of dataset chunkers to keep in per-instance LRU cache."""


class HDF5Root:
    Path = pubroot.PubRoot.Path

    @classmethod
    def get_maker(cls, target: str) -> Callable[[], Self] | None:
        try:
            path = pathlib.Path(target)
            if not h5py.is_hdf5(path):
                return None
        except Exception:
            return None
        return lambda: cls(path)

    def __init__(self, path: pathlib.Path):
        self.h5file = h5py.File(path, mode='r')

    # There must be one cached function per instance,
    # so that it can be reset individually.
    # This means that just ``@functools.(lru_)cache``
    # on e.g. ``_b2args_from_h5dset(self, dset)`` is not enough
    # (there would be a single cache shared by all instances).

    @functools.cached_property
    def _b2args_from_h5dset(self):
        @functools.cache  # TODO: limit size?
        def _getb2args(dset: h5py.Dataset) -> Mapping[str, object]:
            return hdf5.b2args_from_h5dset(dset)
        return _getb2args

    @functools.cached_property
    def _b2attrs_from_h5dset(self):
        @functools.cache  # TODO: limit size?
        def _getb2attrs(dset: h5py.Dataset) -> Mapping[str, object]:
            return hdf5.b2attrs_from_h5dset(dset)
        return _getb2attrs

    @functools.cached_property
    def _b2chunkers_from_h5dset(self):
        @functools.lru_cache(maxsize=_MAX_CACHED_CHUNKERS)  # only hot datasets
        def _getb2chunkers(dset: h5py.Dataset) -> (
                Callable[[int], bytes],
                Callable[[], Iterator[bytes]]):
            b2_args = self._b2args_from_h5dset(dset)
            return hdf5.b2chunkers_from_h5dset(dset, b2_args)
        return _getb2chunkers

    def _clear_caches(self):
        self._b2args_from_h5dset.cache_clear()
        self._b2attrs_from_h5dset.cache_clear()
        self._b2chunkers_from_h5dset.cache_clear()

    def walk_dsets(self) -> Iterator[Path]:
        # TODO: either iterate (without accumulation) or cache
        dsets = []
        warn = logging.getLogger().warning

        def visitor(name, node):
            if not _is_dataset(node):
                if not isinstance(node, h5py.Group):
                    warn("skipping incompatible HDF5 dataset: %r", name)
                return
            # TODO: handle array / frame / (compressed) file distinctly
            dsets.append(self.Path(f'{name}.b2nd'))

        self.h5file.visititems(visitor)
        yield from iter(dsets)

    def _path_to_dset(self, relpath: Path) -> h5py.Dataset:
        name = re.sub(r'\.b2(nd|frame)?$', '', str(relpath))
        node = self.h5file.get(name)
        if node is None or not _is_dataset(node):
            raise pubroot.NoSuchDatasetError(relpath)
        return node

    def exists_dset(self, relpath: Path) -> bool:
        try:
            self._path_to_dset(relpath)
        except pubroot.NoSuchDatasetError:
            return False
        return True

    def get_dset_etag(self, relpath: Path) -> str:
        dset = self._path_to_dset(relpath)
        # All datasets have the modification time of their file.
        h5path = pathlib.Path(self.h5file.filename)
        stat = h5path.stat()
        return f'{stat.st_mtime}:{dset.nbytes}'

    def get_dset_meta(self, relpath: Path) -> pydantic.BaseModel:
        dset = self._path_to_dset(relpath)
        b2_args = self._b2args_from_h5dset(dset)
        b2_attrs = self._b2attrs_from_h5dset(dset)
        b2_array = hdf5.b2uninit_from_h5dset(dset, b2_args, b2_attrs)
        return srv_utils.read_metadata(b2_array)

    def get_dset_chunk(self, relpath: Path, nchunk: int) -> bytes:
        dset = self._path_to_dset(relpath)
        b2getchunk, _ = self._b2chunkers_from_h5dset(dset)
        try:
            return b2getchunk(nchunk)
        except IndexError as ie:
            raise pubroot.NoSuchChunkError(*ie.args) from ie

    def open_dset_raw(self, relpath: Path) -> io.RawIOBase:
        # TODO: handle array / frame / (compressed) file distinctly
        raise NotImplementedError(
            "cannot read raw contents of array dataset")

    async def awatch_dsets(self) -> AsyncIterator[Collection[Path]]:
        h5path = self.h5file.filename
        old_dsets = set(self.walk_dsets())
        async for _ in watchfiles.awatch(h5path):
            self._clear_caches()
            self.h5file.close()
            self.h5file = h5py.File(h5path, mode='r')
            # All datasets are supposed to change along with their file.
            cur_dsets = set(self.walk_dsets())
            # Old datasets are included in case any of them disappeared.
            yield old_dsets | cur_dsets
            old_dsets = cur_dsets


pubroot.register_root_class(HDF5Root)


def _is_dataset(node: h5py.Group | h5py.Dataset) -> bool:
    return isinstance(node, h5py.Dataset) and hdf5.h5dset_is_compatible(node)


def create_example_root(path):
    """Create an example HDF5 file to be used as a root."""
    import h5py
    import numpy
    from hdf5plugin import Blosc2 as B2Comp

    with h5py.File(path, 'x') as h5f:
        h5f.create_dataset('/scalar', data=123.456)
        h5f.create_dataset('/string', data=numpy.bytes_("Hello world!"))

        a = numpy.arange(100, dtype='uint8')
        h5f.create_dataset('/arrays/1d-raw', data=a)

        a = numpy.array([b'foobar'] * 100)
        h5f.create_dataset('/arrays/1ds-blosc2', data=a, chunks=(50,),
                           **B2Comp())

        a = numpy.arange(100, dtype='complex128').reshape(10, 10)
        a = a + a*1j
        h5f.create_dataset('/arrays/2d-nochunks', data=a, chunks=None)

        a = numpy.arange(100, dtype='complex128').reshape(10, 10)
        a = a + a*1j
        h5f.create_dataset('/arrays/2d-gzip', data=a, chunks=(4, 4),
                           compression='gzip')

        a = numpy.arange(1000, dtype='uint8').reshape(10, 10, 10)
        h5f.create_dataset('/arrays/3d-blosc2', data=a, chunks=(4, 10, 10),
                           **B2Comp(cname='lz4', clevel=7,
                                    filters=B2Comp.BITSHUFFLE))

        ds = h5f.create_dataset('/attrs', data=0)
        a = numpy.arange(4, dtype='uint8').reshape(2, 2)
        for k, v in {'Int': 42, 'IntT': numpy.int16(42),
                         'Bin': b"foo", 'BinT': numpy.bytes_(b"foo"),
                         'Str': "bar", # StrT=numpy.str_("bar"),
                         'Arr': a.tolist(), 'ArrT': a,
                         'NilBin': h5py.Empty('|S4'),
                         # NilStr=h5py.Empty('|U4'),
                         'NilInt': h5py.Empty('uint8')}.items():
            ds.attrs[k] = v

        h5f.create_dataset('/unsupported/empty', data=h5py.Empty('float64'))
        h5f.create_dataset('/unsupported/vlstring', data="Hello world!")

        a = numpy.arange(1, dtype='uint8').reshape((1,) * 23)
        h5f.create_dataset('/unsupported/too-many-dimensions', data=a)

        h5f.create_dataset('/unsupported/array-dtype',
                           dtype=numpy.dtype(('float64', (4,))),
                           shape=(10,))

        h5f.create_dataset('/unsupported/compound-dtype',
                           dtype=numpy.dtype('uint8,float64'),
                           shape=(10,))

        h5f['/unsupported/soft-link'] = h5py.SoftLink('/arrays/1d-raw')


def main():
    import os
    import sys
    try:
        _, h5fpath = sys.argv
    except ValueError:
        prog = os.path.basename(sys.argv[0])
        print(f"Usage: {prog} HDF5_FILE", file=sys.stderr)
        sys.exit(1)
    create_example_root(h5fpath)
    print(f"Created example HDF5 root: {h5fpath!r}", file=sys.stderr)


if __name__ == '__main__':
    main()



import platform
import sys

import httpx
import numpy as np
import pytest

import caterva2 as cat2

from .conf import configuration  # noqa: F401
from .files import examples_dir, examples_hdf5  # noqa: F401
from .services import services  # noqa: F401
from .sub_auth import sub_user  # noqa: F401

try:  # Python-Blosc2 is optional
    import blosc2
except ImportError:
    blosc2 = None


def pytest_configure(config):
    print("\n" + "-=" * 38)
    print(f"Caterva2 version:      {cat2.__version__}")
    if blosc2 is not None:
        print(f"Python-Blosc2 version: {blosc2.__version__}")
    print(f"HTTPX version:         {httpx.__version__}")
    print(f"NumPy version:         {np.__version__}")
    print(f"Python version:        {sys.version}")
    print(f"Platform:              {platform.platform()}")
    print(f"Rootdir:               {config.rootdir}")
    print("-=" * 38)


@pytest.fixture(scope="session")
def client(services):  # noqa: F811
    urlbase = services.get_urlbase("subscriber")
    return cat2.Client(urlbase)


@pytest.fixture(scope="session")
def auth_client(services, sub_user):  # noqa: F811
    if not sub_user:
        return None

    urlbase = services.get_urlbase("subscriber")
    return cat2.Client(urlbase, sub_user)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################
import os
import pathlib
import re
import sys

# Requirements
import httpx
import requests

# Caterva2
from caterva2 import utils

# Optional requirements
try:
    import blosc2

    blosc2_is_here = True
except ImportError:
    blosc2_is_here = False


USE_REQUESTS = os.environ.get("USE_REQUESTS", False)
if sys.platform == "emscripten":
    USE_REQUESTS = True


def slice_to_string(slice_):
    if slice_ is None or slice_ == () or slice_ == slice(None):
        return ""
    slice_parts = []
    if not isinstance(slice_, tuple):
        slice_ = (slice_,)
    for index in slice_:
        if isinstance(index, int):
            slice_parts.append(str(index))
        elif isinstance(index, slice):
            start = index.start or ""
            stop = index.stop or ""
            if index.step not in (1, None):
                raise IndexError("Only step=1 is supported")
            # step = index.step or ''
            slice_parts.append(f"{start}:{stop}")
    return ", ".join(slice_parts)


def parse_slice(string):
    if not string:
        return None
    obj = []
    for segment in string.split(","):
        if ":" not in segment:
            segment = int(segment)
        else:
            segment = slice(*(int(x.strip()) if x.strip() else None for x in segment.split(":")))
        obj.append(segment)

    return tuple(obj) if len(obj) > 1 else obj[0]


def get_auth_cookie(urlbase, user_auth):
    """
    Authenticate to a subscriber as a user and get an authorization cookie.

    Authentication fields will usually be ``username`` and ``password``.

    Parameters
    ----------
    urlbase : str
        The base of URLs of the subscriber to query.
    user_auth : dict
        A mapping of fields and values used as data to be posted for
        authenticating the user.

    Returns
    -------
    str
        An authentication token that may be used as a cookie in further
        requests to the subscriber.

    Examples
    --------
    >>> import caterva2 as cat2
    >>> urlbase = 'https://cat2.cloud/demo'
    >>> auth_cookie = cat2.get_auth_cookie(urlbase, dict(username='user@example.com', password='foo'))
    >>> cat2.upload('root-example/ds-sc-attr.b2nd', '@personal/attr.b2nd', urlbase, auth_cookie)
    PosixPath('@personal/attr.b2nd')
    """
    url = f"{urlbase}/auth/jwt/login"
    client, url = get_client_and_url(None, url)

    if hasattr(user_auth, "_asdict"):  # named tuple (from tests)
        user_auth = user_auth._asdict()
    resp = client.post(url, data=user_auth)
    resp.raise_for_status()
    return "=".join(list(resp.cookies.items())[0])


def fetch_data(path, urlbase, params, auth_cookie=None, as_blosc2=False):
    response = _xget(f"{urlbase}/api/fetch/{path}", params=params, auth_cookie=auth_cookie)
    data = response.content
    # Try different deserialization methods
    try:
        data = blosc2.ndarray_from_cframe(data)
    except RuntimeError:
        data = blosc2.schunk_from_cframe(data)
    if as_blosc2:
        return data
    if hasattr(data, "ndim"):  # if b2nd or b2frame
        # catch 0d case where [:] fails
        return data[()] if data.ndim == 0 else data[:]
    else:
        return data[:]


def get_download_url(path, urlbase):
    return f"{urlbase}/api/fetch/{path}"


def b2_unpack(filepath):
    if not blosc2_is_here:
        return filepath
    schunk = blosc2.open(filepath)
    outfile = filepath.with_suffix("")
    with open(outfile, "wb") as f:
        for i in range(schunk.nchunks):
            data = schunk.decompress_chunk(i)
            f.write(data)
    os.unlink(filepath)
    return outfile


# Not completely RFC6266-compliant, but probably good enough.
_attachment_b2fname_rx = re.compile(r';\s*filename\*?\s*=\s*"([^"]+\.b2)"')


def download_url(url, localpath, try_unpack=True, auth_cookie=None):
    client, url = get_client_and_url(None, url)

    localpath = pathlib.Path(localpath)
    localpath.parent.mkdir(parents=True, exist_ok=True)
    if localpath.is_dir():
        # Get the filename from the URL
        localpath /= url.split("/")[-1]

    headers = {"Cookie": auth_cookie} if auth_cookie else None
    if USE_REQUESTS:
        with client.get(url, headers=headers, stream=True) as r:
            r.raise_for_status()
            # Build the local filepath
            cdisp = r.headers.get("content-disposition", "")
            is_b2 = bool(_attachment_b2fname_rx.findall(cdisp))
            if is_b2:
                # Append '.b2' to the filename
                localpath = localpath.with_suffix(localpath.suffix + ".b2")
            with open(localpath, "wb") as f:
                for data in r.iter_content():
                    f.write(data)

    else:
        with client.stream("GET", url, headers=headers) as r:
            r.raise_for_status()
            # Build the local filepath
            cdisp = r.headers.get("content-disposition", "")
            is_b2 = bool(_attachment_b2fname_rx.findall(cdisp))
            if is_b2:
                # Append '.b2' to the filename
                localpath = localpath.with_suffix(localpath.suffix + ".b2")
            with open(localpath, "wb") as f:
                for data in r.iter_bytes():
                    f.write(data)

    if is_b2 and try_unpack:
        localpath = b2_unpack(localpath)

    return localpath


def upload_file(localpath, remotepath, urlbase, auth_cookie=None):
    client, url = get_client_and_url(None, f"{urlbase}/api/upload/{remotepath}")

    headers = {"Cookie": auth_cookie} if auth_cookie else None
    with open(localpath, "rb") as f:
        response = client.post(url, files={"file": f}, headers=headers)
        response.raise_for_status()
    return pathlib.PurePosixPath(response.json())


#
# HTTP client helpers
#


def get_client_and_url(server, url, return_async_client=False):
    # Use requests instead of httpx (for pyodide)
    if USE_REQUESTS and not return_async_client and server is None:
        client = requests.Session()
        return client, url

    # When NOT called from pyodide
    if return_async_client:
        client_class = httpx.AsyncClient
        transport_class = httpx.AsyncHTTPTransport
    else:
        client_class = httpx.Client
        transport_class = httpx.HTTPTransport

    transport = None
    if server is not None:
        if type(server) is str:
            server = utils.Socket(server)

        assert url.startswith("/"), f'expected absolute path, got "{url}"'
        if server.uds:
            transport = transport_class(uds=server.uds)
            url = f"http://localhost{url}"
        else:
            url = f"http://{server}{url}"

    client = client_class(transport=transport)
    return client, url


def _xget(url, params=None, headers=None, timeout=5, auth_cookie=None, server=None, raise_for_status=True):
    client, url = get_client_and_url(server, url)
    if auth_cookie:
        headers = headers.copy() if headers else {}
        headers["Cookie"] = auth_cookie
    response = client.get(url, params=params, headers=headers, timeout=timeout)
    if raise_for_status:
        response.raise_for_status()
    return response


def get(
    url,
    params=None,
    headers=None,
    timeout=5,
    model=None,
    auth_cookie=None,
    server=None,
    raise_for_status=True,
    return_response=False,
):
    response = _xget(
        url, params, headers, timeout, auth_cookie, server=server, raise_for_status=raise_for_status
    )
    if return_response:
        return response

    json = response.json()
    return json if model is None else model(**json)


def post(url, json=None, auth_cookie=None, server=None):
    client, url = get_client_and_url(server, url)
    headers = {"Cookie": auth_cookie} if auth_cookie else None
    response = client.post(url, json=json, headers=headers)
    response.raise_for_status()
    return response.json()



import functools
import io
import pathlib
import sys
from collections.abc import Sequence

import blosc2
import numpy as np
from blosc2 import NDArray, SChunk

from . import api_utils, utils

sub_urlbase_default = "http://localhost:8002"
"""The default base of URLs provided by the subscriber."""


def _format_paths(urlbase, path=None):
    if urlbase is None and sys.platform == "emscripten":
        urlbase = ""

    if isinstance(urlbase, pathlib.Path):
        urlbase = urlbase.as_posix()
    if urlbase.endswith("/"):
        urlbase = urlbase[:-1]
        urlbase = pathlib.PurePosixPath(urlbase)

    if path is not None:
        # Ensure path is a string before checking startswith
        p = path.as_posix() if hasattr(path, "as_posix") else str(path)
        if p.startswith("/"):
            raise ValueError("path cannot start with a slash")

    return urlbase, path


class Root:
    def __init__(self, client, name):
        """
        Represents a remote repository that can be subscribed to.

        Parameters
        ----------
        client : Client
            The client used to interact with the remote repository.
        name : str
            Name of the root to subscribe to.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client("https://demo.caterva2.net")
        >>> root = client.get("example")
        >>> root.file_list[-3:]
        ['ds-sc-attr.b2nd', 'lung-jpeg2000_10x.b2nd', 'tomo-guess-test.b2nd']
        """
        self.client = client
        urlbase, name = _format_paths(client.urlbase, name)
        self.name = name

    @property
    def urlbase(self):
        return self.client.urlbase

    @property
    def cookie(self):
        return self.client.cookie

    @property
    def file_list(self):
        """
        Retrieves a list of files in this root.
        """
        return api_utils.get(f"{self.urlbase}/api/list/{self.name}", auth_cookie=self.cookie)

    def __repr__(self):
        return f"<Root: {self.name}>"

    def __getitem__(self, path):
        """
        Retrieves a file or dataset from the root.

        Parameters
        ----------
        path : str or Path
            Path of the file or dataset to retrieve.

        Returns
        -------
        File, Dataset
            An instance of :class:`File` or :class:`Dataset`.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> root['ds-1d.b2nd']
        <Dataset: example/ds-1d.b2nd>
        """
        path = path.as_posix() if hasattr(path, "as_posix") else str(path)
        if path.endswith((".b2nd", ".b2frame")):
            return Dataset(self, path)
        else:
            return File(self, path)

    def __contains__(self, path):
        """
        Checks if a path exists in the root.

        Parameters
        ----------
        path : str or Path
            The path of the file or dataset to check.

        Returns
        -------
        bool
            True if the path exists in the root, otherwise False.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> 'ds-1d.b2nd' in root
        True
        """
        path = path.as_posix() if hasattr(path, "as_posix") else str(path)
        return path in self.file_list

    def __iter__(self):
        """
        Iterates over the files and datasets in the root.

        Returns
        -------
        iter
            An iterator for the files and datasets in the root.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> for file in root:
        ...     print(file)
        README.md
        dir1/ds-2d.b2nd
        dir1/ds-3d.b2nd
        dir2/ds-4d.b2nd
        ds-1d-b.b2nd
        ds-1d-fields.b2nd
        ds-1d.b2nd
        ds-2d-fields.b2nd
        ds-hello.b2frame
        ds-sc-attr.b2nd
        lung-jpeg2000_10x.b2nd
        tomo-guess-test.b2nd
        """
        return iter(self.file_list)

    def __len__(self):
        """
        Returns the number of files in the root.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> len(root)
        12
        """
        return len(self.file_list)

    def __str__(self):
        return self.name

    def upload(self, localpath, remotepath=None):
        """
        Uploads a local file to this root.

        Parameters
        ----------
        localpath : Path
            Path of the local file to upload.
        remotepath : Path, optional
            Remote path where the file will be uploaded.  If not provided, the
            file will be uploaded to the top level of this root.

        Returns
        -------
        File
            A instance of :class:`File` or :class:`Dataset`.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To upload a file you must be registered as a user.
        >>> client = cat2.Client("https://cat2.cloud/demo", ("joedoe@example.com", "foobar"))
        >>> root = client.get('@personal')
        >>> path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> root.upload('root-example/dir2/ds-4d.b2nd')
        <Dataset: @personal/root-example/dir2/ds-4d.b2nd>
        >>> 'root-example/dir2/ds-4d.b2nd' in root
        True
        """
        if remotepath is None:
            # localpath cannot be absolute in this case (too much prone to errors)
            if pathlib.PurePosixPath(localpath).is_absolute():
                raise ValueError("When `dataset` is not specified, `localpath` must be a relative path")
            remotepath = pathlib.PurePosixPath(self.name) / localpath
        else:
            remotepath = pathlib.PurePosixPath(self.name) / pathlib.PurePosixPath(remotepath)
        uploadpath = self.client.upload(localpath, remotepath)
        # Remove the first component of the upload path (the root name) and return a new File/Dataset
        return self[str(uploadpath.relative_to(self.name))]


class File:
    def __init__(self, root, path):
        """
        Represents a file, which can be a Blosc2 dataset or a regular file on a root repository.

        This class is not intended for direct instantiation; it should be accessed through a
        :class:`Root` instance.

        Parameters
        ----------
        root : Root
            The root repository.
        path : str
            The path of the file.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> file = root['README.md']
        >>> file
        <File: example/README.md>
        >>> file.name
        'README.md'
        >>> file.urlbase
        'https://demo.caterva2.net'
        >>> file.path
        PurePosixPath('example/README.md')
        >>> file.meta['contiguous']
        True
        """
        self.root = root
        _, name = _format_paths(root.urlbase, path)
        _, root = _format_paths(root.urlbase, root.name)
        self.name = name
        self.path = pathlib.PurePosixPath(f"{self.root}/{self.name}")
        self.meta = api_utils.get(f"{self.urlbase}/api/info/{self.path}", auth_cookie=self.cookie)
        # TODO: 'cparams' is not always present (e.g. for .b2nd files)
        # print(f"self.meta: {self.meta['cparams']}")

    @property
    def client(self):
        return self.root.client

    @property
    def urlbase(self):
        return self.client.urlbase

    @property
    def cookie(self):
        return self.client.cookie

    def __repr__(self):
        return f"<File: {self.path}>"

    @functools.cached_property
    def vlmeta(self):
        """
        Returns a mapping of metalayer names to their respective values.

        This is used to access variable-length metalayers (user attributes)
        associated with the file.

        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> file = root['ds-sc-attr.b2nd']
        >>> file.vlmeta
        {'a': 1, 'b': 'foo', 'c': 123.456}
        """
        schunk_meta = self.meta.get("schunk", self.meta)
        return schunk_meta.get("vlmeta", {})

    def get_download_url(self):
        """
        Retrieves the download URL for the file.

        Returns
        -------
        str
            The file's download URL.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> file = root['ds-1d.b2nd']
        >>> file.get_download_url()
        'https://demo.caterva2.net/api/fetch/example/ds-1d.b2nd'
        """
        return api_utils.get_download_url(self.path, self.urlbase)

    def __getitem__(self, key):
        """
        Retrieves a slice of the dataset.

        Parameters
        ----------
        key : int, slice, tuple of ints and slices, or None
            Specifies the slice to fetch.

        Returns
        -------
        numpy.ndarray
            The requested slice of the dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> ds = root['ds-1d.b2nd']
        >>> ds[1]
        array(1)
        >>> ds[:1]
        array([0])
        >>> ds[0:10]
        array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        """
        return self.slice(key, as_blosc2=False)

    def slice(
        self, key: int | slice | Sequence[slice], as_blosc2: bool = True
    ) -> NDArray | SChunk | np.ndarray:
        """Get a slice of a File/Dataset.

        Parameters
        ----------
        key : int, slice, or sequence of slices
            The slice to retrieve.  If a single slice is provided, it will be
            applied to the first dimension.  If a sequence of slices is
            provided, each slice will be applied to the corresponding
            dimension.
        as_blosc2 : bool
            If True (default), the result will be returned as a Blosc2 object
            (either a `SChunk` or `NDArray`).  If False, it will be returned
            as a NumPy array (equivalent to `self[key]`).

        Returns
        -------
        NDArray or SChunk or numpy.ndarray
            A new Blosc2 object containing the requested slice.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> ds = root['ds-1d.b2nd']
        >>> ds.slice(1)
        <blosc2.ndarray.NDArray object at 0x10747efd0>
        >>> ds.slice(1)[()]
        array(1)
        >>> ds.slice(slice(0, 10))[:]
        array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        """
        # Convert slices to strings
        slice_ = api_utils.slice_to_string(key)
        # Fetch and return the data as a Blosc2 object / NumPy array
        return api_utils.fetch_data(
            self.path, self.urlbase, {"slice_": slice_}, auth_cookie=self.cookie, as_blosc2=as_blosc2
        )

    def download(self, localpath=None):
        """
        Downloads the file to storage.

        Parameters
        ----------
        localpath : Path, optional
            The destination path for the downloaded file.  If not specified, the file will
            be downloaded to the current working directory.

        Returns
        -------
        Path
            The path to the downloaded file.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> file = root['ds-1d.b2nd']
        >>> file.download()
        PosixPath('example/ds-1d.b2nd')
        >>> file.download('mydir/myarray.b2nd')
        PosixPath('mydir/myarray.b2nd')
        """
        return self.client.download(
            self.path,
            localpath=localpath,
        )

    def move(self, dst):
        """
        Moves the file to a new location.

        Parameters
        ----------
        dst : Path
            The destination path for the file.

        Returns
        -------
        Path
            The new path of the file after the move.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> # For moving a file you need to be a registered user
        >>> client = cat2.Client("https://cat2.cloud/demo", ("joedoe@example.com", "foobar"))
        >>> root = client.get('@personal')
        >>> root.upload('root-example/dir2/ds-4d.b2nd')
        <Dataset: @personal/root-example/dir2/ds-4d.b2nd>
        >>> file = root['root-example/dir2/ds-4d.b2nd']
        >>> file.move('@personal/root-example/dir1/ds-4d-moved.b2nd')
        PurePosixPath('@personal/root-example/dir1/ds-4d-moved.b2nd')
        >>> 'root-example/dir2/ds-4d.b2nd' in root
        False
        >>> 'root-example/dir1/ds-4d-moved.b2nd' in root
        True
        """
        return self.client.move(self.path, dst)

    def copy(self, dst):
        """
        Copies the file to a new location.

        Parameters
        ----------
        dst : Path
            The destination path for the file.

        Returns
        -------
        Path
            The new path of the copied file.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # For copying a file you need to be a registered user
        >>> client = cat2.Client("https://cat2.cloud/demo", ("joedoe@example.com", "foobar"))
        >>> root = client.get('@personal')
        >>> root.upload('root-example/dir2/ds-4d.b2nd')
        <Dataset: @personal/root-example/dir2/ds-4d.b2nd>
        >>> file = root['root-example/dir2/ds-4d.b2nd']
        >>> file.copy('@personal/root-example/dir2/ds-4d-copy.b2nd')
        PurePosixPath('@personal/root-example/dir2/ds-4d-copy.b2nd')
        >>> 'root-example/dir2/ds-4d.b2nd' in root
        True
        >>> 'root-example/dir2/ds-4d-copy.b2nd' in root
        True
        """
        return self.client.copy(self.path, dst)

    def remove(self):
        """
        Removes the file from the remote repository.

        Returns
        -------
        str
            The path of the removed file.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To remove a file you need to be a registered user
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> root = client.get('@personal')
        >>> path = 'root-example/dir2/ds-4d.b2nd'
        >>> root.upload(path)
        <Dataset: @personal/root-example/dir2/ds-4d.b2nd>
        >>> file = root[path]
        >>> file.remove()
        '@personal/root-example/dir2/ds-4d.b2nd'
        >>> path in root
        False
        """
        return self.client.remove(self.path)


class Dataset(File):
    def __init__(self, root, path):
        """
        Represents a dataset within a Blosc2 container.

        This class is not intended to be instantiated directly; it should be accessed through a
        :class:`Root` instance.

        Parameters
        ----------
        root : Root
            The root repository.
        path : str
            The path of the dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> ds = root['ds-1d.b2nd']
        >>> ds.dtype
        'int64'
        >>> ds.shape
        (1000,)
        >>> ds.chunks
        (100,)
        >>> ds.blocks
        (10,)
        """
        super().__init__(root, path)

    def __repr__(self):
        # TODO: add more info about dims, types, etc.
        return f"<Dataset: {self.path}>"

    @property
    def dtype(self):
        try:
            return self.meta["dtype"]
        except KeyError as e:
            raise AttributeError("'Dataset' object has no attribute 'dtype'.") from e

    @property
    def shape(self):
        try:
            return tuple(self.meta["shape"])
        except KeyError as e:
            raise AttributeError("'Dataset' object has no attribute 'shape'.") from e

    @property
    def chunks(self):
        try:
            return tuple(self.meta["chunks"])
        except KeyError as e:
            raise AttributeError("'Dataset' object has no attribute 'chunks'.") from e

    @property
    def blocks(self):
        try:
            return tuple(self.meta["blocks"])
        except KeyError as e:
            raise AttributeError("'Dataset' object has no attribute 'blocks'.") from e

    def append(self, data):
        """
        Appends data to the dataset.

        Parameters
        ----------
        data : blosc2.NDArray, numpy.ndarray, sequence
            The data to append to the dataset.

        Returns
        -------
        tuple
            The new shape of the dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To append data to a dataset you need to be a registered user
        >>> client = cat2.Client("https://cat2.cloud/demo", ("joedoe@example.com", "foobar"))
        >>> data = client.copy('@public/examples/ds-1d.b2nd', '@personal/ds-1d.b2nd')
        >>> dataset = client.get('@personal')['ds-1d.b2nd']
        >>> dataset.append([1, 2, 3])
        (1003,)
        """
        return self.client.append(self.path, data)


class BasicAuth:
    """
    Basic authentication for HTTP requests.

    Parameters
    ----------
    username : str, optional
        The username for authentication.
    password : str, optional
        The password for authentication.
    """

    def __init__(self, username=None, password=None):
        self.username = username
        self.password = password


class Client:
    def __init__(self, urlbase, auth=None):
        """
        Creates a client for server in urlbase.

        Parameters
        ----------
        urlbase : str, optional
            Base URL of the subscriber to query. Default to
            :py:obj:`caterva2.sub_urlbase_default`.
        auth : tuple, BasicAuth, optional

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client("https://cat2.cloud/demo")
        >>> auth_client = cat2.Client("https://cat2.cloud/demo", ("joedoe@example.com", "foobar"))
        """
        self.urlbase = utils.urlbase_type(urlbase)
        self.cookie = None
        if auth is not None:
            if isinstance(auth, BasicAuth):
                username = auth.username
                password = auth.password
            elif isinstance(auth, tuple):
                username = auth[0]
                password = auth[1]
            else:
                raise ValueError("auth must be BasicAuth or a tuple (username, password)")
            if username and password:
                self.cookie = api_utils.get_auth_cookie(
                    self.urlbase, {"username": username, "password": password}
                )

    def get_roots(self):
        """
        Retrieves the list of available roots.

        Returns
        -------
        dict
            Dictionary mapping available root names to their details:
            - ``name``: the root name
            - ``http``: the HTTP endpoint
            - ``subscribed``: whether it is subscribed or not.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> roots_dict = client.get_roots()
        >>> sorted(roots_dict.keys())
        ['@public', 'b2tests', 'example', 'h5example', 'h5lung_j2k', 'h5numbers_j2k']
        >>> client.subscribe('b2tests')
        'Ok'
        >>> roots_dict['b2tests']
        {'name': 'b2tests', 'http': 'localhost:8014', 'subscribed': True}
        """
        urlbase, _ = _format_paths(self.urlbase)
        return api_utils.get(f"{self.urlbase}/api/roots", auth_cookie=self.cookie)

    def _get_root(self, name):
        """
        Retrieves a specified root name.

        Parameters
        ----------
        name : str
            Name of the root to retrieve.

        Returns
        -------
        Root
            An instance of :class:`Root`.

        """
        if "/" in name:
            raise ValueError("Root names cannot contain slashes")
        # It is a root, subscribe to it
        ret = self.subscribe(name)
        if ret != "Ok":
            roots = self.get_roots()
            raise ValueError(f"Could not subscribe to root {name} (only {roots.keys()} available)")
        return Root(self, name)

    def get(self, path):
        """
        Returns an object for the given path.

        Parameters
        ----------
        path : Path
            Path to the root, file or dataset.

        Returns
        -------
        Object : Root, File, Dataset
            Object representing the root, file or dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root = client.get('example')
        >>> root.name
        'example'
        >>> file = client.get('example/README.md')
        >>> file.name
        'README.md'
        >>> ds = client.get('example/ds-1d.b2nd')
        >>> ds[:10]
        array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
        """
        # Normalize the path to a POSIX path
        path = pathlib.PurePosixPath(path).as_posix()
        # Check if the path is a root or a file/dataset
        if "/" not in path:
            return self._get_root(path)
        # If not a root, assume it's a file/dataset
        root_name = path.split("/")[0]
        root = self._get_root(root_name)
        file_path = path[len(root_name) + 1 :]
        return root[file_path]

    def subscribe(self, root):
        """
        Subscribes to a specified root.

        Parameters
        ----------
        root : str
            Name of the root to subscribe to.

        Returns
        -------
        str
            Server response as a string.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> root_name = 'h5numbers_j2k'
        >>> client.subscribe(root_name)
        'Ok'
        >>> client.get_roots()[root_name]
        {'name': 'h5numbers_j2k', 'http': 'localhost:8011', 'subscribed': True}
        """
        urlbase, root = _format_paths(self.urlbase, root)
        return api_utils.post(f"{self.urlbase}/api/subscribe/{root}", auth_cookie=self.cookie)

    def get_list(self, path):
        """
        Lists datasets in a specified path.

        Parameters
        ----------
        path : str
            Path to a root, directory or dataset.

        Returns
        -------
        list
            List of dataset names as strings, relative to the specified path.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> client.subscribe('example')
        'Ok'
        >>> client.get_list('example')[:3]
        ['README.md', 'dir1/ds-2d.b2nd', 'dir1/ds-3d.b2nd']
        """
        urlbase, path = _format_paths(self.urlbase, path)
        return api_utils.get(f"{self.urlbase}/api/list/{path}", auth_cookie=self.cookie)

    def get_info(self, path):
        """
        Retrieves information about a specified dataset.

        Parameters
        ----------
        path : str
            Path to the dataset.

        Returns
        -------
        dict
            Dictionary of dataset properties, mapping property names to their values.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> client.subscribe('example')
        'Ok'
        >>> path = 'example/ds-2d-fields.b2nd'
        >>> info = client.get_info(path)
        >>> info.keys()
        dict_keys(['shape', 'chunks', 'blocks', 'dtype', 'schunk', 'mtime'])
        >>> info['shape']
        [100, 200]
        """
        urlbase, path = _format_paths(self.urlbase, path)
        return api_utils.get(f"{self.urlbase}/api/info/{path}", auth_cookie=self.cookie)

    def fetch(self, path, slice_=None):
        """
        Retrieves the entire content (or a specified slice) of a dataset.

        Parameters
        ----------
        path : str
            Path to the dataset.
        urlbase : str, optional
            Base URL to query. Defaults to
            :py:obj:`caterva2.sub_urlbase_default`.
        slice_ : int, slice, tuple of ints and slices, or None
            Specifies the slice to fetch. If None, the whole dataset is fetched.

        Returns
        -------
        numpy.ndarray
            The requested slice of the dataset as a Numpy array.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> client.fetch('example/ds-2d-fields.b2nd', (slice(0, 2), slice(0, 2))
        array([[(0.0000000e+00, 1.       ), (5.0002502e-05, 1.00005  )],
               [(1.0000500e-02, 1.0100005), (1.0050503e-02, 1.0100505)]],
              dtype=[('a', '<f4'), ('b', '<f8')])
        """
        urlbase, path = _format_paths(self.urlbase, path)
        slice_ = api_utils.slice_to_string(slice_)  # convert to string
        return api_utils.fetch_data(path, urlbase, {"slice_": slice_}, auth_cookie=self.cookie)

    def get_chunk(self, path, nchunk):
        """
        Retrieves a specified compressed chunk from a file.

        Parameters
        ----------
        path : str
            Path of the dataset.
        nchunk : int
            ID of the unidimensional chunk to retrieve.

        Returns
        -------
        bytes obj
            The compressed chunk data.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> client.subscribe('example')
        'Ok'
        >>> info_schunk = client.get_info('example/ds-2d-fields.b2nd')['schunk']
        >>> info_schunk['nchunks']
        1
        >>> info_schunk['cratio']
        6.453000645300064
        >>> chunk = client.get_chunk('example/ds-2d-fields.b2nd', 0)
        >>> info_schunk['chunksize'] / len(chunk)
        6.453000645300064
        """
        urlbase, path = _format_paths(self.urlbase, path)
        data = api_utils._xget(
            f"{self.urlbase}/api/chunk/{path}", {"nchunk": nchunk}, auth_cookie=self.cookie
        )
        return data.content

    def download(self, dataset, localpath=None):
        """
        Downloads a dataset to local storage.

        **Note:** If the dataset is a regular file and Blosc2 is installed,
        it will be downloaded and decompressed. Otherwise, it will remain
        compressed in its `.b2` format.

        Parameters
        ----------
        dataset : Path
            Path to the dataset.
        localpath : Path, optional
            Local path to save the downloaded dataset. Defaults to the current
            working directory if not specified.

        Returns
        -------
        Path
            The path to the downloaded file.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> path = 'example/ds-2d-fields.b2nd'
        >>> client = cat2.Client('https://demo.caterva2.net')
        >>> client.download(path)
        PosixPath('example/ds-2d-fields.b2nd')
        """
        urlbase, dataset = _format_paths(self.urlbase, dataset)
        url = api_utils.get_download_url(dataset, urlbase)
        localpath = pathlib.Path(localpath) if localpath else None
        if localpath is None:
            path = "." / pathlib.Path(dataset)
        elif localpath.is_dir():
            path = localpath / dataset.name
        else:
            path = localpath
        return api_utils.download_url(
            url, str(path), try_unpack=api_utils.blosc2_is_here, auth_cookie=self.cookie
        )

    def upload(self, localpath, dataset):
        """
        Uploads a local dataset to a remote repository.

        **Note:** If `localpath` is a regular file without a `.b2nd`,
        `.b2frame` or `.b2` extension, it will be automatically compressed
        with Blosc2 on the server, adding a `.b2` extension internally.

        Parameters
        ----------
        localpath : Path
            Path to the local dataset.
        dataset : Path
            Remote path to upload the dataset to.

        Returns
        -------
        Path
            Path of the uploaded file on the server.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To upload a file you need to be authenticated as an already registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> newpath = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> uploaded_path = client.upload('root-example/dir2/ds-4d.b2nd', newpath)
        >>> str(uploaded_path) == newpath
        True
        """
        urlbase, dataset = _format_paths(self.urlbase, dataset)
        return api_utils.upload_file(
            localpath,
            dataset,
            urlbase,
            auth_cookie=self.cookie,
        )

    def append(self, remotepath, data):
        """
        Appends data to the remote location.

        Parameters
        ----------
        remotepath : Path
            Remote path of the dataset to enlarge.
        data : blosc2.NDArray, np.ndarray, sequence
            The data to append.

        Returns
        -------
        tuple
            The new shape of the dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To upload a file you need to be authenticated as an already registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> path = '@personal/ds-1d.b2nd'
        >>> client.copy('@public/examples/ds-1d.b2nd', path)
        PurePosixPath('@personal/ds-1d.b2nd')
        >>> ndarray = blosc2.arange(0, 10)
        >>> client.append(path, ndarray)
        (1010,)
        """
        if not hasattr(data, "shape"):
            array = np.asarray(data)
        else:
            array = data

        ndarray = blosc2.asarray(array)
        cframe = ndarray.to_cframe()
        file = io.BytesIO(cframe)

        client, url = api_utils.get_client_and_url(None, f"{self.urlbase}/api/append/{remotepath}")
        headers = {"Cookie": self.cookie}
        response = client.post(url, files={"file": file}, headers=headers)
        response.raise_for_status()
        return tuple(response.json())

    def remove(self, path):
        """
        Removes a dataset or the contents of a directory from a remote repository.

        **Note:** When a directory is removed, only its contents are deleted;
        the directory itself remains. This behavior allows for future
        uploads to the same directory. It is subject to in future versions.

        Parameters
        ----------
        path : Path
            Path of the dataset or directory to remove.

        Returns
        -------
        str
            The path that was removed.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To remove a file you need to be a registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> uploaded_path = client.upload('root-example/dir2/ds-4d.b2nd', path)
        >>> removed_path = client.remove(path)
        >>> removed_path == path
        True
        """
        urlbase, path = _format_paths(self.urlbase, path)
        return api_utils.post(f"{self.urlbase}/api/remove/{path}", auth_cookie=self.cookie)

    def move(self, src, dst):
        """
        Moves a dataset or directory to a new location.

        Parameters
        ----------
        src : Path
            Source path of the dataset or directory.
        dst : Path
            The destination path for the dataset or directory.

        Returns
        -------
        str
            New path of the moved dataset or directory.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To move a file you need to be a registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> uploaded_path = client.upload('root-example/dir2/ds-4d.b2nd', path)
        >>> newpath = f'@personal/dir{np.random.randint(0, 100)}/ds-4d-moved.b2nd'
        >>> moved_path = client.move(path, newpath)
        >>> str(moved_path) == newpath
        True
        >>> path.replace('@personal/', '') in client.get_list('@personal')
        False
        """
        urlbase, _ = _format_paths(self.urlbase)
        result = api_utils.post(
            f"{self.urlbase}/api/move/",
            {"src": str(src), "dst": str(dst)},
            auth_cookie=self.cookie,
        )
        return pathlib.PurePosixPath(result)

    def copy(self, src, dst):
        """
        Copies a dataset or directory to a new location.

        Parameters
        ----------
        src : Path
            Source path of the dataset or directory.
        dst : Path
            Destination path for the dataset or directory.

        Returns
        -------
        str
            New path of the copied dataset or directory.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To copy a file you need to be a registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> src_path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> uploaded_path = client.upload('root-example/dir2/ds-4d.b2nd', src_path)
        >>> copy_path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d-copy.b2nd'
        >>> copied_path = client.copy(src_path, copy_path)
        >>> str(copied_path) == copy_path
        True
        >>> datasets = client.get_list('@personal')
        >>> src_path.replace('@personal/', '') in datasets
        True
        >>> copy_path.replace('@personal/', '') in datasets
        True
        """
        urlbase, _ = _format_paths(self.urlbase)
        result = api_utils.post(
            f"{self.urlbase}/api/copy/", {"src": str(src), "dst": str(dst)}, auth_cookie=self.cookie
        )
        return pathlib.PurePosixPath(result)

    def lazyexpr(self, name, expression, operands=None, compute=False):
        """
        Creates a lazy expression dataset in personal space.

        A dataset with the specified name will be created or overwritten if already
        exists.

        Parameters
        ----------
        name : str
            Name of the dataset to be created (without extension).
        expression : str
            Expression to be evaluated, which must yield a lazy expression.
        operands : dict
            Mapping of variables in the expression to their corresponding dataset paths.
        compute : bool, optional
            If false, generate lazyexpr and do not compute anything.
            If true, compute lazy expression on creation and save (full) result.
            Default false.

        Returns
        -------
        Path
            Path of the created dataset.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To create a lazyexpr you need to be a registered used
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> src_path = f'@personal/dir{np.random.randint(0, 100)}/ds-4d.b2nd'
        >>> path = client.upload('root-example/dir1/ds-2d.b2nd', src_path)
        >>> client.lazyexpr('example-expr', 'a + a', {'a': path})
        PurePosixPath('@personal/example-expr.b2nd')
        >>> 'example-expr.b2nd' in client.get_list('@personal')
        True
        """
        urlbase, _ = _format_paths(self.urlbase)
        # Convert possible Path objects in operands to strings so that they can be serialized
        if operands is not None:
            operands = {k: str(v) for k, v in operands.items()}
        else:
            operands = {}
        expr = {"name": name, "expression": expression, "operands": operands, "compute": compute}
        dataset = api_utils.post(f"{self.urlbase}/api/lazyexpr/", expr, auth_cookie=self.cookie)
        return pathlib.PurePosixPath(dataset)

    def adduser(self, newuser, password=None, superuser=False):
        """
        Adds a user to the subscriber.

        Parameters
        ----------
        newuser : str
            Username of the user to add.
        password : str, optional
            Password for the user to add.
        superuser : bool, optional
            Indicates if the user is a superuser.

        Returns
        -------
        str
            An explanatory message about the operation's success or failure.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To add a user you need to be a superuser
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> username = f'user{np.random.randint(0, 100)}@example.com'
        >>> message = client.adduser(username, 'foo')
        >>> f"User added: username='{username}' password='foo' superuser=False" == message
        True
        """
        urlbase, _ = _format_paths(self.urlbase)
        return api_utils.post(
            f"{self.urlbase}/api/adduser/",
            {"username": newuser, "password": password, "superuser": superuser},
            auth_cookie=self.cookie,
        )

    def deluser(self, user):
        """
        Deletes a user from the subscriber.

        Parameters
        ----------
        username : str
            Username of the user to delete.

        Returns
        -------
        str
            An explanatory message about the operation's success or failure.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To delete a user you need to be a superuser
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> username = f'user{np.random.randint(0, 100)}@example.com'
        >>> _ = client.adduser(username, 'foo')
        >>> message = client.deluser(username)
        >>> message == f"User deleted: {username}"
        True
        """
        urlbase, _ = _format_paths(self.urlbase)
        return api_utils.get(f"{self.urlbase}/api/deluser/{user}", auth_cookie=self.cookie)

    def listusers(self, username=None):
        """
        Lists the users in the subscriber.

        Parameters
        ----------
        username : str, optional
            Username of the specific user to list.

        Returns
        -------
        list of dict
            A list of user dictionaries in the subscriber.

        Examples
        --------
        >>> import caterva2 as cat2
        >>> import numpy as np
        >>> # To list the users you need to be a superuser
        >>> client = cat2.Client('https://cat2.cloud/demo', ("joedoe@example.com", "foobar"))
        >>> users = client.listusers()
        >>> sorted(users[0].keys())
        ['email', 'hashed_password', 'id', 'is_active', 'is_superuser', 'is_verified']
        >>> username = f'user{np.random.randint(0, 100)}@example.com'
        >>> _ = client.adduser(username, 'foo')
        >>> updated_users = client.listusers()
        >>> len(users) + 1 == len(updated_users)
        True
        >>> user_info = client.listusers(username)
        >>> user_info[0]['is_superuser']
        False
        >>> superuser_info = client.listusers('superuser@example.com')
        >>> superuser_info[0]['is_superuser']
        True
        """
        urlbase, _ = _format_paths(self.urlbase)
        url = f"{self.urlbase}/api/listusers/" + (f"?username={username}" if username else "")
        return api_utils.get(url, auth_cookie=self.cookie)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'Caterva2'
copyright = '2024, ironArray SLU'
author = 'ironArray SLU'
release = '0.3.dev0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosummary",
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "numpydoc",
    "myst_parser",
    "sphinx_paramlinks",
    "sphinx_design",
    "nbsphinx",
]

myst_enable_extensions = [
    "html_image",
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_static_path = ['_static']
source_suffix = [".rst", ".md"]
html_theme = "pydata_sphinx_theme"
html_css_files = [
    "css/custom.css",
]
html_favicon = "_static/logo-caterva2-16x16.png"

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "github_url": "https://github.com/ironArray/Caterva2",
    "icon_links": [
        {
            "name": "ironArray Web Site",
            "url": "https://ironarray.io/",
            "icon": "_static/ironArray-icon2-128px.png",
            "type": "local",
        },
    ],
    "collapse_navigation": True,
    "navigation_with_keys": False,
}

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logo-caterva2-horizontal-half.png"



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################
import os

import pytest

from caterva2.services import srv_utils


def sub_auth_enabled():
    return bool(os.environ.get("CATERVA2_SECRET"))


def make_sub_user(services):
    if not sub_auth_enabled():
        return None

    state_dir = services.state_dir / "subscriber"
    return srv_utils.add_user(
        "user@example.com", password="foobar11", is_superuser=True, state_dir=state_dir
    )


@pytest.fixture(scope="session")
def sub_user(services):
    # TODO: This does not work with external services,
    # as their state directory is unknown.
    # One would need to register a new user via the API there.
    return make_sub_user(services)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################
import contextlib
import pathlib

import blosc2
import httpx
import numpy as np
import pytest

import caterva2 as cat2

from .services import TEST_CATERVA2_ROOT, TEST_STATE_DIR

try:
    chdir_ctxt = contextlib.chdir
except AttributeError:  # Python < 3.11
    import os

    @contextlib.contextmanager
    def chdir_ctxt(path):
        cwd = os.getcwd()
        os.chdir(path)
        yield
        os.chdir(cwd)


@pytest.fixture
def pub_host(services):
    return services.get_endpoint(f"publisher.{TEST_CATERVA2_ROOT}")


@pytest.fixture
def fill_public(client, examples_dir):
    # Manually copy some files to the public area (TEST_STATE_DIR)
    fnames = ["README.md", "ds-1d.b2nd", "ds-1d-fields.b2nd", "dir1/ds-2d.b2nd"]
    for fname in fnames:
        orig = examples_dir / fname
        data = orig.read_bytes()
        if not fname.endswith(("b2nd", "b2frame")):
            fname += ".b2"
            schunk = blosc2.SChunk(data=data)
            data = schunk.to_cframe()
        dest = pathlib.Path(TEST_STATE_DIR) / f"subscriber/public/{fname}"
        dest.parent.mkdir(parents=True, exist_ok=True)
        dest.write_bytes(data)
    # We need a user here in case we want to remove files from @public
    return fnames, client.get("@public")


@pytest.fixture
def fill_auth(auth_client, fill_public):
    if not auth_client:
        return None
    fnames, _ = fill_public
    return fnames, auth_client.get("@public")


def test_subscribe(client, auth_client):
    assert client.subscribe(TEST_CATERVA2_ROOT) == "Ok"
    assert client.subscribe("@public") == "Ok"
    for root in ["@personal", "@shared"]:
        if auth_client:
            assert auth_client.subscribe(root) == "Ok"
        else:
            with pytest.raises(Exception) as e_info:
                _ = client.subscribe(root)
            assert "Unauthorized" in str(e_info)


def test_roots(pub_host, client, auth_client):
    client = auth_client if auth_client else client
    roots = client.get_roots()
    assert roots[TEST_CATERVA2_ROOT]["name"] == TEST_CATERVA2_ROOT
    assert roots[TEST_CATERVA2_ROOT]["http"] == pub_host
    assert roots["@public"]["name"] == "@public"
    assert roots["@public"]["http"] == ""
    if auth_client:
        # Special roots (only available when authenticated)
        assert roots["@personal"]["name"] == "@personal"
        assert roots["@personal"]["http"] == ""
        assert roots["@shared"]["name"] == "@shared"
        assert roots["@shared"]["http"] == ""


def test_get_root(client, auth_client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    assert myroot.name == TEST_CATERVA2_ROOT
    assert myroot.urlbase == client.urlbase
    mypublic = client.get("@public")
    assert mypublic.name == "@public"
    assert mypublic.urlbase == client.urlbase
    if auth_client:
        mypersonal = auth_client.get("@personal")
        assert mypersonal.name == "@personal"
        assert mypersonal.urlbase == auth_client.urlbase
        myshared = auth_client.get("@shared")
        assert myshared.name == "@shared"
        assert myshared.urlbase == auth_client.urlbase


def test_get_file(client):
    myfile = client.get(TEST_CATERVA2_ROOT + "/README.md")
    assert myfile.name == "README.md"


def test_get_dataset(client):
    myds = client.get(TEST_CATERVA2_ROOT + "/ds-1d.b2nd")
    assert myds.name == "ds-1d.b2nd"
    assert isinstance(myds, cat2.Dataset)
    assert myds.shape == (1000,)
    assert myds.dtype == np.dtype("int64")
    assert myds.chunks == (100,)
    assert myds.blocks == (10,)
    assert myds.urlbase == client.urlbase


def test_list(client, auth_client, examples_dir):
    myroot = client.get(TEST_CATERVA2_ROOT)
    example = examples_dir
    files = {str(f.relative_to(str(example))) for f in example.rglob("*") if f.is_file()}
    assert set(myroot.file_list) == files
    if auth_client:
        mypersonal = auth_client.get("@personal")
        # In previous tests we have created some files in the personal area
        assert len(mypersonal.file_list) >= 0
        myshared = auth_client.get("@shared")
        assert set(myshared.file_list) == set()


def test_list_public(client, fill_public):
    fnames, mypublic = fill_public
    assert set(mypublic.file_list) == set(fnames)
    # Test toplevel list
    flist = client.get_list("@public")
    assert len(flist) == 4
    for fname in flist:
        assert fname in fnames
    # Test directory list
    flist = client.get_list("@public/dir1")
    assert len(flist) == 1
    for fname in flist:
        assert fname == "ds-2d.b2nd"
    # Test directory list with trailing slash
    flist = client.get_list("@public/dir1/")
    assert len(flist) == 1
    for fname in flist:
        assert fname == "ds-2d.b2nd"
    # Test single dataset list
    flist = client.get_list("@public/dir1/ds-2d.b2nd")
    assert len(flist) == 1
    for fname in flist:
        assert fname == "ds-2d.b2nd"


def test_file(client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    file = myroot["README.md"]
    assert file.name == "README.md"
    assert file.urlbase == client.urlbase


def test_file_public(client, fill_public):
    fnames, mypublic = fill_public
    for fname in fnames:
        file = mypublic[fname]
        assert file.name == fname
        assert file.urlbase == client.urlbase


def test_dataset_info(client, fill_public):
    fnames, mypublic = fill_public
    for fname in fnames:
        if type(mypublic[fname]) is cat2.Dataset:  # Files cannot be expected to have attributes
            info = client.get_info("@public/" + fname)
            data = mypublic[fname]
            assert data.dtype == info["dtype"]
            assert data.shape == tuple(info["shape"])
            assert data.blocks == tuple(info["blocks"])
            assert data.chunks == tuple(info["chunks"])


@pytest.mark.parametrize("dirpath", [None, "dir1", "dir2", "dir2/dir3/dir4"])
@pytest.mark.parametrize("final_dir", [True, False])
def test_move(auth_client, dirpath, final_dir, fill_auth):
    if not auth_client:
        return pytest.skip("authentication support needed")

    fnames, mypublic = fill_auth
    myshared = auth_client.get("@shared")
    for fname in fnames:
        file = mypublic[fname]
        if final_dir:
            new_fname = f"{dirpath}" if dirpath else ""
        else:
            new_fname = f"{dirpath}/{fname}" if dirpath else fname
        newpath = file.move(f"{myshared.name}/{new_fname}")
        assert fname not in mypublic
        if final_dir:
            basename = fname.split("/")[-1]
            new_path = f"{new_fname}/{basename}" if dirpath else basename
            assert str(newpath) == f"{myshared.name}/{new_path}"
            assert myshared[new_path].path == newpath
        else:
            assert str(newpath) == f"{myshared.name}/{new_fname}"
            assert myshared[new_fname].path == newpath
    return None


@pytest.mark.parametrize("dest", ["..", ".", "foo", "foo/bar"])
def test_move_not_allowed(auth_client, dest, fill_auth):
    if not auth_client:
        return pytest.skip("authentication support needed")

    fnames, mypublic = fill_auth
    for fname in fnames:
        # Move the file to a non-special root and check for an exception
        file = mypublic[fname]
        with pytest.raises(Exception) as e_info:
            _ = file.move(dest)
        assert "Bad Request" in str(e_info)
        assert fname in mypublic
    return None


@pytest.mark.parametrize("dirpath", [None, "dir1", "dir2", "dir2/dir3/dir4"])
@pytest.mark.parametrize("final_dir", [True, False])
def test_copy(auth_client, dirpath, final_dir, fill_auth):
    if not auth_client:
        return pytest.skip("authentication support needed")

    fnames, mypublic = fill_auth
    myshared = auth_client.get("@shared")
    for fname in fnames:
        file = mypublic[fname]
        if final_dir:
            new_fname = f"{dirpath}" if dirpath else ""
        else:
            new_fname = f"{dirpath}/{fname}" if dirpath else fname
        newpath = file.copy(f"{myshared.name}/{new_fname}")
        assert fname in mypublic
        if final_dir:
            basename = fname.split("/")[-1]
            new_path = f"{new_fname}/{basename}" if dirpath else basename
            assert str(newpath) == f"{myshared.name}/{new_path}"
            assert myshared[new_path].path == newpath
        else:
            assert str(newpath) == f"{myshared.name}/{new_fname}"
            assert myshared[new_fname].path == newpath
    return None


@pytest.mark.parametrize("fields", [True, False])
def test_append(auth_client, fields, fill_auth, examples_dir):
    if not auth_client:
        return pytest.skip("authentication support needed")

    fnames, mypublic = fill_auth
    myshared = auth_client.get("@shared")
    fname = "ds-1d.b2nd" if not fields else "ds-1d-fields.b2nd"
    # Copy a 1d dataset to the shared area
    file = mypublic[fname]
    newpath = file.copy(f"@shared/{fname}")
    assert newpath == myshared[fname].path
    # Append to the dataset
    if fields:
        data = np.asarray(
            [(1000, 1.0, b"foobar1000", False), (1001, 2.0, b"foobar1001", True)],
            dtype=[("a", "<i4"), ("b", "<f8"), ("c", "S10"), ("d", "?")],
        )
    else:
        data = [1, 2, 3]
    sfile = myshared[fname]
    new_shape = sfile.append(data)
    assert new_shape == (len(data) + file.meta["shape"][0],)

    # Check the data
    fname = examples_dir / fname
    a = blosc2.open(fname)
    b = np.concatenate([a[:], data])
    return np.testing.assert_array_equal(sfile[:], b)


@pytest.mark.parametrize(
    "slice_",
    [1, slice(None, 1), slice(0, 10), slice(10, 20), slice(None), slice(10, 20, 1)],
)
def test_dataset_getitem_fetch(slice_, examples_dir, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["ds-hello.b2frame"]
    assert ds.name == "ds-hello.b2frame"
    assert ds.urlbase == client.urlbase

    example = examples_dir / ds.name
    a = blosc2.open(example)[:]
    if isinstance(slice_, int):
        assert ord(ds[slice_]) == a[slice_]  # TODO: why do we need ord() here?
    else:
        assert ds[slice_] == a[slice_]


def test_dataset_step_diff_1(client):
    # XXX This should fail with anonymous client, because the root is not @public
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["ds-hello.b2frame"]
    assert ds.name == "ds-hello.b2frame"
    assert ds.urlbase == client.urlbase
    # We don't support step != 1
    with pytest.raises(Exception) as e_info:  # noqa: PT012
        _ = ds[::2]
        assert str(e_info.value) == "Only step=1 is supported"


@pytest.mark.parametrize(
    "slice_",
    [0, 1, slice(None, 1), slice(0, 10), slice(10, 20), slice(None), slice(1, 5, 1)],
)
def test_getitem_dataset_1d(slice_, examples_dir, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["ds-1d.b2nd"]
    assert ds.name == "ds-1d.b2nd"
    assert ds.urlbase == client.urlbase

    example = examples_dir / ds.name
    a = blosc2.open(example)[:]
    np.testing.assert_array_equal(ds[slice_], a[slice_])


@pytest.mark.parametrize(
    "slice_",
    [
        1,
        slice(None, 1),
        slice(0, 10),
        slice(10, 20),
        slice(None),
        slice(1, 5, 1),
        (slice(None, 10), slice(None, 20)),
    ],
)
@pytest.mark.parametrize("name", ["dir1/ds-2d.b2nd", "dir2/ds-4d.b2nd"])
def test_getitem_dataset_nd(slice_, name, examples_dir, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot[name]
    example = examples_dir / ds.name
    a = blosc2.open(example)[:]
    np.testing.assert_array_equal(ds[slice_], a[slice_])


@pytest.mark.parametrize(
    "slice_",
    [
        1,
        slice(None, 1),
        slice(0, 10),
        slice(10, 20),
        slice(None),
        slice(1, 5, 1),
        (slice(None, 10), slice(None, 20)),
    ],
)
@pytest.mark.parametrize("name", ["dir1/ds-2d.b2nd", "dir2/ds-4d.b2nd"])
def test_slice_dataset_nd(slice_, name, examples_dir, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot[name]
    example = examples_dir / ds.name
    a = blosc2.open(example)[:]
    arr = ds.slice(slice_)
    assert isinstance(arr, blosc2.NDArray)
    np.testing.assert_array_equal(arr[()], a[slice_])


@pytest.mark.parametrize(
    "slice_",
    [1, slice(None, 1), slice(0, 10), slice(10, 20), slice(None), slice(1, 5, 1)],
)
def test_getitem_regular_file(slice_, examples_dir, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["README.md"]

    # Data contents
    example = examples_dir / ds.name
    a = open(example).read().encode()
    if isinstance(slice_, int):
        assert ord(ds[slice_]) == a[slice_]  # TODO: why do we need ord() here?
    else:
        assert ds[slice_] == a[slice_]


@pytest.mark.parametrize(
    "slice_",
    [1, slice(None, 1), slice(0, 10), slice(10, 20), slice(None), slice(1, 5, 1)],
)
def test_getitem_client_regular_file(slice_, examples_dir, client):
    # Data contents
    example = examples_dir / "README.md"
    a = open(example).read().encode()
    if isinstance(slice_, int):
        assert ord(client.fetch(TEST_CATERVA2_ROOT + "/" + "README.md", slice_=slice_)) == a[slice_]
    else:
        assert client.fetch(TEST_CATERVA2_ROOT + "/" + "README.md", slice_=slice_) == a[slice_]


@pytest.mark.parametrize(
    "slice_",
    [0, 1, slice(None, 1), slice(0, 10), slice(10, 20), slice(None), slice(1, 5, 1)],
)
def test_getitem_client_1d(slice_, examples_dir, client):
    example = examples_dir / "ds-1d.b2nd"
    a = blosc2.open(example)[:]
    np.testing.assert_array_equal(
        client.fetch(TEST_CATERVA2_ROOT + "/" + "ds-1d.b2nd", slice_=slice_), a[slice_]
    )


@pytest.mark.parametrize(
    "slice_",
    [
        1,
        slice(None, 1),
        slice(0, 10),
        slice(10, 20),
        slice(None),
        slice(1, 5, 1),
        (slice(None, 10), slice(None, 20)),
    ],
)
@pytest.mark.parametrize("name", ["dir1/ds-2d.b2nd", "dir2/ds-4d.b2nd"])
def test_getitem_client_nd(slice_, name, examples_dir, client):
    example = examples_dir / name
    a = blosc2.open(example)[:]
    arr = client.fetch(TEST_CATERVA2_ROOT + "/" + name, slice_=slice_)
    np.testing.assert_array_equal(arr, a[slice_])


@pytest.mark.parametrize("name", ["ds-1d.b2nd", "dir1/ds-2d.b2nd"])
def test_download_b2nd(name, examples_dir, tmp_path, client, auth_client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot[name]
    with chdir_ctxt(tmp_path):
        path = ds.download()
        assert path == ds.path

    # Data contents
    example = examples_dir / name
    a = blosc2.open(example)
    with chdir_ctxt(tmp_path):
        b = blosc2.open(path)
        np.testing.assert_array_equal(a[:], b[:])

    # Using 2-step download
    urlpath = ds.get_download_url()
    data = httpx.get(urlpath, headers={"Cookie": auth_client.cookie} if auth_client else None)
    assert data.status_code == 200
    b = blosc2.ndarray_from_cframe(data.content)
    np.testing.assert_array_equal(a[:], b[:])


def test_download_b2frame(examples_dir, tmp_path, client, auth_client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["ds-hello.b2frame"]
    with chdir_ctxt(tmp_path):
        path = ds.download()
        assert path == ds.path

    # Data contents
    example = examples_dir / ds.name
    a = blosc2.open(example)
    with chdir_ctxt(tmp_path):
        b = blosc2.open(path)
        assert a[:] == b[:]

    # Using 2-step download
    urlpath = ds.get_download_url()
    assert urlpath == f"{client.urlbase}/api/fetch/{ds.path}"
    data = httpx.get(urlpath, headers={"Cookie": auth_client.cookie} if auth_client else None)
    assert data.status_code == 200
    b = blosc2.schunk_from_cframe(data.content)
    assert a[:] == b[:]


@pytest.mark.parametrize(
    "fnames",
    [
        ("ds-1d.b2nd", "ds-1d2.b2nd"),
        ("dir1/ds-2d.b2nd", "dir2/ds-2d2.b2nd"),
        ("dir1/ds-2d.b2nd", "dir2/dir3/dir4/ds-2d2.b2nd"),
        ("dir1/ds-2d.b2nd", "dir2/dir3/dir4/"),
    ],
)
def test_download_localpath(fnames, examples_dir, tmp_path, client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    name, localpath = fnames
    ds = myroot[name]
    with chdir_ctxt(tmp_path):
        if localpath.endswith("/"):
            # Create a directory in localpath
            localpath2 = pathlib.Path(localpath)
            localpath2.mkdir(parents=True, exist_ok=True)
        path = ds.download(localpath)
        if localpath.endswith("/"):
            localpath = localpath + name.split("/")[-1]
        assert str(path) == localpath

    # Data contents
    example = examples_dir / name
    a = blosc2.open(example)
    with chdir_ctxt(tmp_path):
        b = blosc2.open(path)
        np.testing.assert_array_equal(a[:], b[:])


def test_download_regular_file(examples_dir, tmp_path, client, auth_client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["README.md"]
    with chdir_ctxt(tmp_path):
        path = ds.download()
        assert path == ds.path

    # Data contents
    example = examples_dir / ds.name
    a = open(example).read()
    with chdir_ctxt(tmp_path):
        b = open(path).read()
        assert a[:] == b[:]

    # Using 2-step download
    urlpath = ds.get_download_url()
    assert urlpath == f"{client.urlbase}/api/fetch/{ds.path}"
    data = httpx.get(urlpath, headers={"Cookie": auth_client.cookie} if auth_client else None)
    assert data.status_code == 200
    b = blosc2.schunk_from_cframe(data.content)
    # TODO: why do we need .decode() here?
    assert a[:] == b[:].decode()


def test_download_public_file(examples_dir, fill_public, tmp_path):
    fnames, mypublic = fill_public
    for fname in fnames:
        # Download the file
        ds = mypublic[fname]
        with chdir_ctxt(tmp_path):
            path = ds.download()
            assert path == ds.path
            # Check data contents
            example = examples_dir / ds.name
            a = open(example, "rb").read()
            b = open(path, "rb").read()
            assert a[:] == b[:]


@pytest.mark.parametrize(
    "fnames",
    [
        ("ds-1d.b2nd", None),
        ("ds-hello.b2frame", None),
        ("README.md", None),
        ("README.md", "README2.md"),
        ("dir1/ds-2d.b2nd", None),
        ("dir1/ds-2d.b2nd", "dir2/ds-2d.b2nd"),
        ("dir1/ds-2d.b2nd", "dir2/dir3/dir4/ds-2d2.b2nd"),
        ("dir1/ds-3d.b2nd", "dir2/dir3/dir4/"),
    ],
)
@pytest.mark.parametrize("root", ["@personal", "@shared", "@public"])
@pytest.mark.parametrize("remove", [False, True])
def test_upload(fnames, remove, root, examples_dir, tmp_path, auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    localpath, remotepath = fnames
    remote_root = auth_client.get(root)
    myroot = auth_client.get(TEST_CATERVA2_ROOT)
    ds = myroot[localpath]
    with chdir_ctxt(tmp_path):
        path = ds.download()
        assert path == ds.path
        # Check whether path exists and is a file
        assert path.exists()
        assert path.is_file()
        # Now, upload the file to the remote root
        remote_ds = remote_root.upload(path, remotepath)
        # Check whether the file has been uploaded with the correct name
        if remotepath:
            if remotepath.endswith("/"):
                assert remote_ds.name == remotepath + path.name
            else:
                assert remote_ds.name == remotepath
        else:
            assert remote_ds.name == str(path)
        # Check removing the file
        if remove:
            remote_removed = pathlib.Path(remote_ds.remove())
            assert remote_removed == remote_ds.path
            # Check that the file has been removed
            with pytest.raises(Exception) as e_info:
                _ = remote_root[remote_removed]
            assert "Not Found" in str(e_info.value)


def test_upload_public_unauthorized(client, auth_client, examples_dir, tmp_path):
    if auth_client:
        pytest.skip("not authentication needed")

    remote_root = client.get("@public")
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["README.md"]
    with chdir_ctxt(tmp_path):
        path = ds.download()
        assert path == ds.path
        with pytest.raises(Exception) as e_info:
            _ = remote_root.upload(path)
        assert "Unauthorized" in str(e_info)


@pytest.mark.parametrize("name", ["ds-1d.b2nd", "ds-hello.b2frame", "README.md"])
def test_vlmeta(client, name):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot[name]
    schunk_meta = ds.meta.get("schunk", ds.meta)
    assert ds.vlmeta is schunk_meta["vlmeta"]


def test_vlmeta_data(client):
    myroot = client.get(TEST_CATERVA2_ROOT)
    ds = myroot["ds-sc-attr.b2nd"]
    assert ds.vlmeta == {"a": 1, "b": "foo", "c": 123.456}


### Lazy expressions


def test_lazyexpr(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    opnm = "ds"
    oppt = f"{TEST_CATERVA2_ROOT}/ds-1d.b2nd"
    expression = f"{opnm} + 0"
    operands = {opnm: oppt}
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)
    opinfo = auth_client.get_info(oppt)
    lxpath = auth_client.lazyexpr(lxname, expression, operands)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")

    # Check result metadata.
    lxinfo = auth_client.get_info(lxpath)
    assert lxinfo["shape"] == opinfo["shape"]
    assert lxinfo["dtype"] == opinfo["dtype"]
    assert lxinfo["expression"] == f"({expression})"
    assert lxinfo["operands"] == operands

    # Check result data.
    a = auth_client.fetch(oppt)
    b = auth_client.fetch(lxpath)
    np.testing.assert_array_equal(a[:], b[:])


def test_lazyexpr_getchunk(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    opnm = "ds"
    oppt = f"{TEST_CATERVA2_ROOT}/ds-1d.b2nd"
    expression = f"{opnm} - 0"
    operands = {opnm: oppt}
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)
    lxpath = auth_client.lazyexpr(lxname, expression, operands)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")

    # Check for chunksize and dtype
    opinfo = auth_client.get_info(oppt)
    chunksize = opinfo["chunks"][0]
    dtype = opinfo["dtype"]

    # Get the first chunks
    chunk_ds = auth_client.get_chunk(oppt, 0)
    chunk_expr = auth_client.get_chunk(lxpath, 0)
    # Check data
    out = np.empty(chunksize, dtype=dtype)
    blosc2.decompress2(chunk_ds, out)
    out_expr = np.empty(chunksize, dtype=dtype)
    blosc2.decompress2(chunk_expr, out_expr)
    np.testing.assert_array_equal(out, out_expr)


def test_expr_from_expr(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    opnm = "ds"
    oppt = f"{TEST_CATERVA2_ROOT}/ds-1d.b2nd"
    expression = f"{opnm} + 1"
    operands = {opnm: oppt}
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)
    opinfo = auth_client.get_info(oppt)
    lxpath = auth_client.lazyexpr(lxname, expression, operands)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")

    expression2 = f"{opnm} * 2"
    operands2 = {opnm: f"@personal/{lxname}.b2nd"}
    lxname = "expr_from_expr"
    lxpath2 = auth_client.lazyexpr(lxname, expression2, operands2)
    assert lxpath2 == pathlib.Path(f"@personal/{lxname}.b2nd")

    # Check result metadata.
    lxinfo = auth_client.get_info(lxpath)
    lxinfo2 = auth_client.get_info(lxpath2)
    assert lxinfo["shape"] == opinfo["shape"] == lxinfo2["shape"]
    assert lxinfo["dtype"] == opinfo["dtype"] == lxinfo2["dtype"]
    assert lxinfo["expression"] == f"({expression})"
    assert lxinfo2["expression"] == f"({expression2})"
    assert lxinfo["operands"] == operands
    assert lxinfo2["operands"] == operands2

    # Check result data.
    a = auth_client.fetch(oppt)
    b = auth_client.fetch(lxpath)
    c = auth_client.fetch(lxpath2)
    np.testing.assert_array_equal(a[:] + 1, b[:])
    np.testing.assert_array_equal((a[:] + 1) * 2, c[:])


def test_expr_no_operand(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    expression = "linspace(0, 10)"
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)
    lxpath = auth_client.lazyexpr(lxname, expression)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")
    c = auth_client.get(lxpath)
    a = blosc2.linspace(0, 10)
    np.testing.assert_array_equal(a[:], c[:])

    # Check error when operand should be present but isn't
    opnm = "ds"
    oppt = f"{TEST_CATERVA2_ROOT}/ds-1d.b2nd"
    expression = "ds + linspace(0, 10)"
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)
    with pytest.raises(Exception) as e_info:
        lxpath = auth_client.lazyexpr(lxname, expression)


def test_expr_force_compute(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    expression = "linspace(0, 10)"
    lxname = "my_expr"

    auth_client.subscribe(TEST_CATERVA2_ROOT)

    # Uncomputed lazyexpr is a blosc2 lazyexpr
    lxpath = auth_client.lazyexpr(lxname, expression, compute=False)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")
    c = auth_client.get(lxpath)
    assert c.meta["expression"] == expression

    # Computed lazyexpr is a blosc2 array
    lxpath = auth_client.lazyexpr(lxname, expression, compute=True)
    assert lxpath == pathlib.Path(f"@personal/{lxname}.b2nd")
    c = auth_client.get(lxpath)
    assert c.meta.get("expression", None) is None


# User management
def test_adduser(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    username = "test@user.com"
    password = "testpassword"
    is_superuser = False
    message = auth_client.adduser(username, password, is_superuser)
    assert "User added" in message
    lusers = auth_client.listusers()
    assert username in [user["email"] for user in lusers]
    # Delete the user for future tests
    message = auth_client.deluser(username)
    assert "User deleted" in message


def test_adduser_malformed(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    username = "test_noat_user"
    password = "testpassword"
    is_superuser = False
    with pytest.raises(Exception) as e_info:
        _ = auth_client.adduser(username, password, is_superuser)
    assert "Bad Request" in str(e_info)


def test_adduser_maxexceeded(auth_client, configuration):
    if not auth_client:
        pytest.skip("authentication support needed")

    # TODO: make this to work; currently this returns None
    # maxusers = configuration.get("subscriber.maxusers")
    # For now, keep in sync with subscriber.maxusers in caterva2/tests/caterva2-login.toml
    maxusers = 5
    # Add maxusers users; we already have one user, so the next loop should fail
    # when reaching the creation of last user
    for n in range(maxusers):
        # This should work fine for n < maxusers
        username = f"test{n}@user.com"
        password = "testpassword"
        is_superuser = False
        if n == maxusers - 1:  # we already have one user
            with pytest.raises(Exception) as e_info:
                _ = auth_client.adduser(username, password, is_superuser)
            assert "Bad Request" in str(e_info.value)
        else:
            message = auth_client.adduser(username, password, is_superuser)
            assert "User added" in message
    # Remove the created users
    for m in range(n):
        username = f"test{m}@user.com"
        message = auth_client.deluser(username)
        assert "User deleted" in message
    # Count the current number of users
    data = auth_client.listusers()
    assert len(data) == 1


def test_adduser_unauthorized(client, auth_client):
    if auth_client:
        pytest.skip("not authentication needed")

    username = "test@user.com"
    password = "testpassword"
    is_superuser = False
    with pytest.raises(Exception) as e_info:
        _ = client.adduser(username, password, is_superuser)
    assert "Not Found" in str(e_info)


def test_deluser(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    username = "test@user.com"
    password = "testpassword"
    is_superuser = False
    message = auth_client.adduser(username, password, is_superuser)
    assert "User added" in message
    # Now, delete the user
    message = auth_client.deluser(username)
    assert "User deleted" in message
    # Check that the user has been deleted
    with pytest.raises(Exception) as e_info:
        _ = auth_client.deluser(username)
    assert "Bad Request" in str(e_info)


def test_deluser_unauthorized(client, auth_client):
    if auth_client:
        pytest.skip("not authentication needed")

    username = "test@user.com"
    with pytest.raises(Exception) as e_info:
        _ = client.deluser(username)
    assert "Not Found" in str(e_info)


def test_listusers(auth_client):
    if not auth_client:
        pytest.skip("authentication support needed")

    username = "test@user.com"
    password = "testpassword"
    is_superuser = False
    message = auth_client.adduser(username, password, is_superuser)
    assert "User added" in message
    # List users
    data = auth_client.listusers()
    assert username in [user["email"] for user in data]
    # Delete the user
    message = auth_client.deluser(username)
    assert "User deleted" in message
    # List users again
    data = auth_client.listusers()
    assert username not in [user["email"] for user in data]


def test_listusers_unauthorized(client, auth_client):
    if auth_client:
        pytest.skip("not authentication needed")

    with pytest.raises(Exception) as e_info:
        _ = client.listusers()
    assert "Not Found" in str(e_info)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Subscriber user management.

Used for user authentication (for the moment), based on FastAPI Users example:
https://fastapi-users.github.io/fastapi-users/latest/configuration/full-example/

The database should only be used if a (non-empty) secret token for the
management of users is set in the 'CATERVA2_SECRET' environment variable.
"""

import functools
import os
import uuid

from fastapi import Depends, Request
from fastapi_mail import ConnectionConfig, FastMail, MessageSchema, MessageType
from fastapi_users import BaseUserManager, FastAPIUsers, InvalidPasswordException, UUIDIDMixin
from fastapi_users.authentication import (
    AuthenticationBackend,
    CookieTransport,
    JWTStrategy,
)
from fastapi_users.db import SQLAlchemyUserDatabase

from .db import User, get_user_db

conf = ConnectionConfig(
    MAIL_USERNAME="",
    MAIL_PASSWORD="",
    MAIL_FROM="noreply@cat2.cloud",
    MAIL_PORT=25,
    MAIL_SERVER="localhost",
    MAIL_STARTTLS=False,
    MAIL_SSL_TLS=False,
    USE_CREDENTIALS=False,
    VALIDATE_CERTS=False,
    # SUPRESS_SEND=True,
    MAIL_DEBUG=True,
)


async def send_email(recipients, body):
    from caterva2.services import settings

    message = MessageSchema(
        subject="Reset password of your cat2.cloud account",
        recipients=recipients,
        body=body,
        subtype=MessageType.html,
    )

    if settings.urlbase.startswith("http://localhost:"):
        print("<<<<<<<<<<")
        print(message.body)
        print(">>>>>>>>>>")
    else:
        fm = FastMail(conf)
        await fm.send_message(message)


class UserManager(UUIDIDMixin, BaseUserManager[User, uuid.UUID]):
    @functools.cached_property
    def reset_password_token_secret(self):
        return os.environ.get("CATERVA2_SECRET")

    @functools.cached_property
    def verification_token_secret(self):
        return os.environ.get("CATERVA2_SECRET")

    # TODO: Replace with actual functionality;
    # support user verification and user deletion.

    async def validate_password(self, password: str, user):
        if len(password) < 8:
            raise InvalidPasswordException(reason="Password should be at least 8 characters")

    async def on_after_register(self, user: User, request: Request | None = None):
        print(f"User {user.email} with id {user.id} has been added.")

    async def on_after_forgot_password(self, user: User, token: str, request: Request | None = None):
        from caterva2.services.sub import make_url, templates

        template = templates.get_template("emails/forgot-password.html")
        url = make_url(request, "html-reset-password", token=token)
        body = template.render({"reset_url": url})
        await send_email([user.email], body)

    async def on_after_request_verify(self, user: User, token: str, request: Request | None = None):
        print(f"Verification requested for user {user.id}. Verification token: {token}")


async def get_user_manager(user_db: SQLAlchemyUserDatabase = Depends(get_user_db)):
    yield UserManager(user_db)


cookie_transport = CookieTransport(
    cookie_name="c2subauth",
    cookie_secure=False,  # TODO: only for testing
)


def get_jwt_strategy() -> JWTStrategy:
    # The token itself is valid for an hour, even after an explicit logout
    # (however the cookie transport would delete the cookie anyway).
    lifetime_seconds = 3600 * 12
    return JWTStrategy(secret=os.environ.get("CATERVA2_SECRET"), lifetime_seconds=lifetime_seconds)


auth_backend = AuthenticationBackend(
    name="jwt",
    transport=cookie_transport,
    get_strategy=get_jwt_strategy,
)

fastapi_users = FastAPIUsers[User, uuid.UUID](get_user_manager, [auth_backend])

current_active_user = fastapi_users.current_user(
    active=True,
    verified=False,  # TODO: set when verification works
)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import asyncio
import collections
import contextlib
import json
import pathlib
import random
import string
import typing

# Requirements
import blosc2
import fastapi
import fastapi_websocket_pubsub
import safer
import uvicorn
from fastapi_users.exceptions import UserNotExists
from sqlalchemy.future import select

# Project
from caterva2 import models
from caterva2.services.subscriber import db as sub_db
from caterva2.services.subscriber import schemas as sub_schemas
from caterva2.services.subscriber import users as sub_users


def compress_file(path):
    with open(path, "rb") as src:
        data = src.read()
        schunk = blosc2.SChunk(data=data)
        data = schunk.to_cframe()
        path2 = f"{path}.b2"

    with open(path2, "wb") as dst:
        dst.write(data)

    path.unlink()


def cache_lookup(cachedir, path, may_not_exist=False):
    if cachedir == path:
        # Special case for the cache root
        return path
    path = pathlib.Path(path)
    if (cachedir / path).is_dir():
        # Special case for directories
        return cachedir / path

    if path.suffix not in {".b2frame", ".b2nd"} and not may_not_exist:
        if path.is_file():
            compress_file(path)
        path = f"{path}.b2"

    return get_abspath(cachedir, path, may_not_exist)


def get_model_from_obj(obj, model_class, **kwargs):
    if isinstance(obj, dict):

        def getter(o, k):
            return o[k]
    else:
        getter = getattr

    data = kwargs.copy()
    for key, info in model_class.model_fields.items():
        if key not in data:
            try:
                value = getter(obj, key)
            except AttributeError:
                continue

            if info.annotation is str:
                value = str(value)

            data[key] = value

    return model_class(**data)


def read_metadata(obj, cache=None, personal=None, shared=None, public=None):
    # Open dataset
    if isinstance(obj, pathlib.Path):
        path = obj
        if not path.is_file():
            raise FileNotFoundError(f'File "{path}" does not exist or is a directory')

        assert path.suffix in {".b2frame", ".b2nd", ".b2"}
        obj = blosc2.open(path)
        stat = path.stat()
        mtime = stat.st_mtime
    else:
        mtime = None

    # Read metadata
    if isinstance(obj, blosc2.ndarray.NDArray):
        array = obj
        cparams = get_model_from_obj(array.schunk.cparams, models.CParams)
        cparams = reformat_cparams(cparams)
        schunk = get_model_from_obj(array.schunk, models.SChunk, cparams=cparams)
        return get_model_from_obj(array, models.Metadata, schunk=schunk, mtime=mtime)
    elif isinstance(obj, blosc2.schunk.SChunk):
        schunk = obj
        cparams = get_model_from_obj(schunk.cparams, models.CParams)
        cparams = reformat_cparams(cparams)
        return get_model_from_obj(schunk, models.SChunk, cparams=cparams, mtime=mtime)
    elif isinstance(obj, blosc2.LazyExpr):
        operands = operands_as_paths(obj.operands, cache, personal, shared, public)
        return get_model_from_obj(obj, models.LazyArray, operands=operands, mtime=mtime)
    else:
        raise TypeError(f"unexpected {type(obj)}")


def reformat_cparams(cparams):
    cparams.__setattr__(
        "filters, meta",
        [
            (cparams.filters[i], cparams.filters_meta[i])
            for i in range(len(cparams.filters))
            if cparams.filters[i] != blosc2.Filter.NOFILTER
        ],
    )
    #   delattr(cparams, 'filters')
    #   delattr(cparams, 'filters_meta')
    return cparams


def get_relpath(ndarr, cache, personal, shared, public):
    path = pathlib.Path(ndarr.schunk.urlpath)
    if shared is not None and path.is_relative_to(shared):
        # Shared: /.../<shared>/<subpath> to <path> (i.e. no change)
        return path
    elif public is not None and path.is_relative_to(public):
        # Shared: /.../<public>/<subpath> to <path> (i.e. no change)
        return path
    try:
        # Cache: /.../<root>/<subpath> to <root>/<subpath>
        path = path.relative_to(cache)
    except ValueError:
        # personal: /.../<uid>/<subpath> to @personal/<subpath>
        path = path.relative_to(personal)
        parts = list(path.parts)
        parts[0] = "@personal"
        path = pathlib.Path(*parts)

    return path


def operands_as_paths(operands, cache, personal, shared, public):
    return {
        nm: None if op is None else str(get_relpath(op, cache, personal, shared, public))
        for (nm, op) in operands.items()
    }


#
# Pub/Sub helpers
#


def start_client(url):
    client = fastapi_websocket_pubsub.PubSubClient()
    client.start_client(url)
    return client


async def disconnect_client(client, timeout=5):
    if client is not None:
        # If the broker is down client.disconnect hangs, wo we wrap it in a timeout
        await asyncio.wait_for(client.disconnect(), timeout)


#
# HTTP server helpers
#

HeaderType = typing.Annotated[str | None, fastapi.Header()]


def raise_bad_request(detail):
    raise fastapi.HTTPException(status_code=400, detail=detail)


def raise_unauthorized(detail="Unauthorized"):
    raise fastapi.HTTPException(status_code=401, detail=detail)


def raise_not_found(detail="Not Found"):
    raise fastapi.HTTPException(status_code=404, detail=detail)


def get_abspath(root, path, may_not_exist=False):
    abspath = root / path

    # Security check
    if root not in abspath.parents:
        raise_bad_request(f"Invalid path {path}")

    # Existence check
    if not abspath.is_file() and not may_not_exist:
        raise_not_found()

    return abspath


def check_dset_path(proot, path):
    try:
        exists = proot.exists_dset(path)
    except ValueError:
        raise_bad_request(f"Invalid path {path}")
    else:
        if not exists:
            raise_not_found()


def uvicorn_run(app, args, root_path=""):
    http = args.http
    if http.uds:
        uvicorn.run(app, uds=http.uds, root_path=root_path)
    else:
        uvicorn.run(app, host=http.host, port=http.port, root_path=root_path)


#
# Blosc2 related helpers
#


def compress(data, dst=None):
    assert isinstance(data, (bytes, pathlib.Path))

    if dst is not None:
        dst.parent.mkdir(exist_ok=True, parents=True)
        if dst.exists():
            dst.unlink()

    # Create schunk
    cparams = {}
    dparams = {}
    storage = {
        "urlpath": dst,
        "cparams": cparams,
        "dparams": dparams,
    }
    schunk = blosc2.SChunk(**storage)

    # Append data
    if isinstance(data, pathlib.Path):
        with open(data, "rb") as f:
            data = f.read()

    schunk.append_data(data)

    return schunk


def iterchunk(data: bytes, chunk_size=2**20):
    for i in range(0, len(data), chunk_size):
        yield data[i : i + chunk_size]


#
# Facility to persist program state
#


class Database:
    def __init__(self, path, initial):
        self.path = path
        self.model = initial.__class__
        if path.exists():
            self.load()
        else:
            path.parent.mkdir(exist_ok=True, parents=True)
            self.data = initial
            self.save()

    def load(self):
        with self.path.open() as file:
            dump = json.load(file)
            self.data = self.model.model_validate(dump)

    def save(self):
        dump = self.data.model_dump_json(exclude_none=True)
        with safer.open(self.path, "w") as file:
            file.write(dump)

    def __getattr__(self, name):
        return getattr(self.data, name)


#
# Subscriber related
#

# <https://fastapi-users.github.io/fastapi-users/10.3/cookbook/create-user-programmatically/>
UserAuth = collections.namedtuple("UserAuth", ["username", "password"])


async def aadd_user(username, password, is_superuser, state_dir=None):
    if password is None:
        password = "".join([random.choice(string.ascii_letters) for i in range(8)])
    user = UserAuth(username=username, password=password)

    sub_state = state_dir
    sub_state.mkdir(parents=True, exist_ok=True)
    await sub_db.create_db_and_tables(sub_state)
    cx = contextlib.asynccontextmanager
    async with (
        cx(sub_db.get_async_session)() as session,
        cx(sub_db.get_user_db)(session) as udb,
        cx(sub_users.get_user_manager)(udb) as umgr,
    ):
        # Check that the user does not exist
        try:
            await umgr.get_by_email(user.username)
            return user
        except UserNotExists:
            schema = sub_schemas.UserCreate(
                email=user.username, password=user.password, is_superuser=is_superuser
            )
            await umgr.create(schema)

    return user


def add_user(username, password, is_superuser, state_dir=None):
    return asyncio.run(aadd_user(username, password, is_superuser, state_dir=state_dir))


async def adel_user(username: str):
    async with (
        contextlib.asynccontextmanager(sub_db.get_async_session)() as session,
        contextlib.asynccontextmanager(sub_db.get_user_db)(session) as udb,
        contextlib.asynccontextmanager(sub_users.get_user_manager)(udb) as umgr,
    ):
        user = await umgr.get_by_email(username)
        if user:
            await umgr.delete(user)


def del_user(username):
    return asyncio.run(adel_user(username))


async def alist_users(username=None):
    async with (
        contextlib.asynccontextmanager(sub_db.get_async_session)() as session,
        contextlib.asynccontextmanager(sub_db.get_user_db)(session) as udb,
    ):
        query = select(udb.user_table)
        if username:
            query = query.where(udb.user_table.email == username)
        result = await session.execute(query)
        # Return a list of dictionaries
        return result.scalars().all()


def list_users(username=None):
    return asyncio.run(alist_users(username))



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

import argparse
import contextlib
import datetime
import logging
import os
import pathlib

try:
    import tomllib as toml
except ImportError:
    import tomli as toml

#
# Context managers
#


@contextlib.contextmanager
def log_exception(logger, message):
    try:
        yield
    except Exception:
        logger.exception(message)


#
# Datetime related
#


def epoch_to_iso(time):
    return datetime.datetime.fromtimestamp(time, tz=datetime.UTC).isoformat()


#
# Filesystem helpers
#


def walk_files(root, exclude=None):
    if exclude is None:
        exclude = set()

    if root is not None:
        for path in root.glob("**/*"):
            if path.is_file():
                relpath = path.relative_to(root)
                if str(relpath) not in exclude:
                    yield path, relpath


def iterdir(root):
    for path in root.iterdir():
        relpath = path.relative_to(root)
        yield path, relpath


#
# Command line helpers
#
def urlbase_type(string):
    if not (string.startswith("http://") or string.startswith("https://")):
        raise ValueError("invalid HTTP(S) URL base", string)
    return string


class Socket(str):
    host = None
    port = None
    uds = None

    def __init__(self, string):
        try:
            host, port = string.rsplit(":", maxsplit=1)
        except ValueError:
            host = port = None

        if host and port:
            # Host and port
            self.host = host
            self.port = int(port)
            # Remove brackets from IPv6 addresses
            if host.startswith("[") and host.endswith("]"):
                self.host = host[1:-1]
        else:
            # Unix domain socket
            self.uds = string


def get_parser(loglevel="warning", statedir=None, id=None, http=None, url=None, broker=None):
    parser = argparse.ArgumentParser()
    _add_preliminary_args(parser, id=id)  # just for help purposes
    if broker is not None:
        parser.add_argument("--broker", default=broker, type=Socket, help="socket address of the broker")
    if http is not None:
        parser.add_argument(
            "--http", default=http, type=Socket, help="Listen to given hostname:port or unix socket"
        )
    if statedir is not None:
        parser.add_argument("--statedir", default=statedir, type=pathlib.Path)
    parser.add_argument("--loglevel", default=loglevel)
    return parser


def run_parser(parser):
    args = parser.parse_args()

    # Logging
    loglevel = args.loglevel.upper()
    logging.basicConfig(level=loglevel)

    return args


#
# Configuration file
#

conf_file_name = "caterva2.toml"


class Conf:
    def __init__(self, conf, prefix=None, id=None):
        self._conf = conf
        self.prefix = prefix
        self.id = id

        self._pfx = (f"{prefix}.{id}" if id else prefix) if prefix else None

    def get(self, key, default=None):
        """Get configuration item with dot-separated `key`.

        The `default` value is returned if any part of the `key` is missing.

        The prefix (if given) is prepended to a `key` that starts with a dot.
        If an ID was given, it is appended to `prefix`, separated by a dot.
        """
        if self._pfx is not None and key.startswith("."):
            key = self._pfx + key
        keys = key.split(".")
        conf = self._conf
        for k in keys:
            conf = conf.get(k)
            if conf is None:
                return default
        return conf


def _add_preliminary_args(parser, id=None):
    parser.add_argument(
        "--conf",
        default=conf_file_name,
        type=pathlib.Path,
        help=("path to alternative configuration file " "(may not exist)"),
    )
    if id is not None:  # the empty string is a valid (default) ID
        parser.add_argument(
            "--id", default=id, help=("a string to distinguish services " "of the same category")
        )


def _parse_preliminary_args(allow_id):
    parser = argparse.ArgumentParser(add_help=False)
    _add_preliminary_args(parser, id="" if allow_id else None)
    return parser.parse_known_args()[0]


def get_conf(prefix=None, allow_id=False):
    """Get settings from the configuration file, if existing.

    If the configuration file does not exist, return an empty configuration.

    You may get the value for a key from the returned configuration ``conf``
    with ``conf.get('path.to.item'[, default])``.  If a `prefix` is given and
    the key starts with a dot, like ``.path.to.item``, `prefix` is prepended
    to it.  If `allow_id` is true and command line arguments has a non-empty
    value for the ``--id`` option, the value gets appended to `prefix`,
    separated by a dot.

    For instance, with ``conf = get_conf('foo')`` and ``--id=bar``,
    ``conf.get('.item')`` is equivalent to ``conf.get('foo.bar.item')``.
    """
    opts = _parse_preliminary_args(allow_id)
    if allow_id and opts.id and any(p in opts.id for p in [os.curdir, os.pardir, os.sep]):
        raise ValueError("invalid identifier", opts.id)

    id_ = opts.id if allow_id else None
    try:
        with open(opts.conf, "rb") as conf_file:
            conf = toml.load(conf_file)
            return Conf(conf, prefix=prefix, id=id_)
    except FileNotFoundError:
        return Conf({}, prefix=prefix, id=id_)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""Subscriber schemas.

Used for user authentication (for the moment), based on FastAPI Users example:
https://fastapi-users.github.io/fastapi-users/latest/configuration/full-example/
"""

import uuid

from fastapi_users import schemas


class UserRead(schemas.BaseUser[uuid.UUID]):
    pass


class UserCreate(schemas.BaseUserCreate):
    pass


class UserUpdate(schemas.BaseUserUpdate):
    pass



"""Caterva2 services for tests.

This ensures that Caterva2 broker, publisher and subscriber services are
running before proceeding to tests.  It has three modes of operation:

- Standalone script: when run as a script, it starts the services as children
  and makes sure that they are available to other local programs.  If given an
  argument, it uses it as the directory to store state in; otherwise it uses
  the value in `DEFAULT_STATE_DIR`.  If the directory does not exist, it is
  created and populated with example datasets.  If further arguments are
  given, each one is taken as a root description ``[ROOT_NAME=]ROOT_SOURCE``
  with an optional root name (`TEST_DEFAULT_ROOT` by default) and the source
  of example datasets to be copied.

  Terminating the program stops the services.

  Usage example::

      $ cd Caterva2
      $ python -m caterva2.tests.services &  # state in ``_caterva2``
      [3] 12345
      $ pytest
      $ kill %3

- pytest fixture with external services: when using `services()` as a fixture,
  it checks that the services are available to other local programs.  It does
  not tamper with the state directory nor stop the services when tests finish.

  Usage example: same as above (but on the pytest side).

- pytest fixture with managed services: if the environment variable
  ``CATERVA2_USE_EXTERNAL`` is set to 1, the `services()` fixture uses
  external services; otherwise, it takes care of starting the services as
  children and making sure that they are available to other local programs.
  It also uses the value in `TEST_STATE_DIR` as the directory to store state
  in.  If the directory exists, it is removed first. Then the directory is
  created and populated with the example files from the source distribution.
  When tests finish, the services are stopped.

  Usage example::

      $ cd Caterva2
      $ env CATERVA2_USE_EXTERNAL=1 pytest  # state in ``_caterva2_tests``
"""

import collections
import functools
import itertools
import logging
import os
import pathlib
import re
import shutil
import signal
import subprocess
import sys
import time
from pathlib import Path

import httpx
import pytest

BASE_DIR = pathlib.Path(__file__).resolve().parent
DEFAULT_STATE_DIR = "_caterva2"
TEST_STATE_DIR = DEFAULT_STATE_DIR + "_tests"
TEST_DEFAULT_ROOT = "foo"
TEST_CATERVA2_ROOT = TEST_DEFAULT_ROOT
TEST_HDF5_ROOT = "hdf5root"

local_port_iter = itertools.count(8100)
logger = logging.getLogger("tests")


def service_ep_getter(first):
    def get_service_ep():
        nonlocal first
        if first is not None:
            ep, first = first, None
            return ep
        return f"localhost:{next(local_port_iter)}"

    return get_service_ep


get_bro_ep = service_ep_getter("localhost:8000")
get_pub_ep = service_ep_getter("localhost:8001")
get_sub_ep = service_ep_getter("localhost:8002")


def make_get_http(host, path="/"):
    def check():
        url = f"http://{host}{path}"
        try:
            r = httpx.get(url, timeout=0.5)
            return 100 <= r.status_code < 500  # not internal errors
        except httpx.ConnectError:
            return False

    check.__name__ = f"get_http_{host}"  # more descriptive
    check.host = host  # to get final value
    return check


def http_service_check(conf, conf_sect, def_host, path):
    return make_get_http(conf.get(f"{conf_sect}.http", def_host), path)


def bro_check(conf):
    return http_service_check(conf, "broker", get_bro_ep(), "/api/roots")


def pub_check(id_, conf):
    return http_service_check(conf, f"publisher.{id_}", get_pub_ep(), "/api/list")


def sub_check(conf):
    return http_service_check(conf, "subscriber", get_sub_ep(), "/api/roots")


TestRoot = collections.namedtuple("TestRoot", ["name", "source"])


class Services:
    def __init__(self):  # mostly to appease QA
        pass


class ManagedServices(Services):
    def __init__(self, state_dir, reuse_state=True, roots=None, configuration=None):
        super().__init__()

        self.state_dir = Path(state_dir).resolve()
        self.reuse_state = reuse_state
        self.roots = list(roots)
        self.configuration = configuration

        self._procs = {}
        self._endpoints = {}
        self._setup_done = False

    def _start_proc(self, name, *args, check=None):
        if check is not None and check():
            raise RuntimeError(
                f'check for service "{name}" succeeded before start'
                f" (external service running?): {check.__name__}"
            )

        if os.environ.get("CATERVA2_SECRET"):
            conf_file = "caterva2-login.toml"
        else:
            conf_file = "caterva2-nologin.toml"

        popen_args = [
            sys.executable,
            f"-mcaterva2.services.{name[:3]}",
            f"--conf={BASE_DIR / conf_file}",
            f"--statedir={self.state_dir / name}",
            *([f"--http={check.host}"] if check else []),
            *args,
        ]
        popen = subprocess.Popen(popen_args, stdout=sys.stdout, stderr=sys.stderr)
        command_line = " ".join(str(x) for x in popen_args)
        self._procs[name] = popen

        if check is None:
            return

        self._endpoints[name] = check.host

        start_timeout_secs = 10
        start_sleep_secs = 1
        for _ in range(int(start_timeout_secs / start_sleep_secs)):
            returncode = popen.poll()
            if returncode is not None:
                raise RuntimeError(
                    f"service {name} failed with returncode={returncode}, " f"command = {command_line}"
                )
            time.sleep(start_sleep_secs)
            if check():
                break
        else:
            raise RuntimeError(
                f"service {name} timeout after {start_timeout_secs} seconds, " f"command = {command_line}"
            )

    def _get_data_path(self, root):
        return self.state_dir / f"data.{root.name}"

    def _setup(self):
        if self._setup_done:
            return

        if not self.reuse_state and self.state_dir.is_dir():
            shutil.rmtree(self.state_dir)
        self.state_dir.mkdir(exist_ok=True)

        for root in self.roots:
            data_path = self._get_data_path(root)
            if root.source.is_dir():
                if not data_path.exists():
                    shutil.copytree(root.source, data_path, symlinks=True)
                data_path.mkdir(exist_ok=True)
            elif not data_path.exists():
                shutil.copy(root.source, data_path)

        self._setup_done = True

    def start_all(self):
        self._setup()

        self._start_proc("broker", check=bro_check(self.configuration))
        for root in self.roots:
            self._start_proc(
                f"publisher.{root.name}",
                root.name,
                self._get_data_path(root),
                check=pub_check(root.name, self.configuration),
            )
        self._start_proc("subscriber", check=sub_check(self.configuration))

    def stop_all(self):
        for proc in self._procs.values():
            try:
                os.kill(proc.pid, signal.SIGTERM)
                time.sleep(1)
                os.kill(proc.pid, signal.SIGHUP)
            except ProcessLookupError:
                pass

    def wait_for_all(self):
        for proc in self._procs.values():
            proc.wait()

    def get_endpoint(self, service):
        return self._endpoints.get(service)

    def get_urlbase(self, service):
        return f"http://{self.get_endpoint(service)}"


class ExternalServices(Services):
    def __init__(self, roots=None, configuration=None):
        super().__init__()
        self.roots = list(roots)
        self.configuration = conf = configuration

        self._checks = checks = {}
        checks["broker"] = bro_check(conf)
        for root in roots:
            checks[f"publisher.{root.name}"] = pub_check(root.name, conf)
        checks["subscriber"] = sub_check(conf)

    def start_all(self):
        failed = [check.__name__ for check in self._checks.values() if not check()]
        if failed:
            raise RuntimeError("failed checks for external services: " + " ".join(failed))

    def stop_all(self):
        pass

    def wait_for_all(self):
        pass

    def get_endpoint(self, service):
        if service not in self._checks:
            return None
        return self._checks[service].host

    def get_urlbase(self, service):
        ep = self.get_endpoint(service)
        return f"http://{ep}" if ep else None


@pytest.fixture(scope="session")
def services(configuration, examples_dir, examples_hdf5):
    # TODO: Consider using a temporary directory to avoid
    # polluting the current directory with test files
    # and tests being influenced by the presence of a configuration file.
    roots = [TestRoot(TEST_CATERVA2_ROOT, examples_dir)]
    if examples_hdf5 is not None:
        roots.append(TestRoot(TEST_HDF5_ROOT, examples_hdf5))

    srvs = (
        ExternalServices(roots=roots, configuration=configuration)
        if os.environ.get("CATERVA2_USE_EXTERNAL", "0") == "1"
        else ManagedServices(TEST_STATE_DIR, reuse_state=False, roots=roots, configuration=configuration)
    )

    try:
        srvs.start_all()
        yield srvs
    finally:
        srvs.stop_all()
    srvs.wait_for_all()


# Inspired by <https://towerbabbel.com/go-defer-in-python/>.
def defers(func):
    @functools.wraps(func)
    def wrapper(*args, **kwds):
        deferred = []
        try:
            return func(*args, defer=deferred.append, **kwds)
        finally:
            for f in reversed(deferred):
                f()

    return wrapper


@defers
def main(defer):
    from . import conf, files, sub_auth

    roots = [TestRoot(TEST_DEFAULT_ROOT, files.get_examples_dir())]
    hdf5source = files.make_examples_hdf5()
    if hdf5source:
        defer(lambda: (hdf5source.parent.is_dir() and shutil.rmtree(hdf5source.parent)))
        roots.append(TestRoot(TEST_HDF5_ROOT, hdf5source))

    if "--help" in sys.argv:
        rspecs = " ".join(f'"{r.name}={r.source}"' for r in roots)
        print(f"Usage: {sys.argv[0]} " f'[STATE_DIRECTORY="{DEFAULT_STATE_DIR}" [ROOTS={rspecs}]]')
        return

    state_dir = sys.argv[1] if len(sys.argv) >= 2 else DEFAULT_STATE_DIR
    if len(sys.argv) >= 3:
        roots = []  # user-provided roots
    rnames = {r.name for r in roots}
    rarg_rx = re.compile(r"(?:(.+?)=)?(.+)")  # ``[name=]source``
    for rarg in sys.argv[2:]:
        rname, rsource = rarg_rx.match(rarg).groups()
        rname = rname if rname else TEST_DEFAULT_ROOT
        if rname in rnames:
            raise ValueError(
                f"root name {rname!r} already in use; " f"please set a different name for {rsource!r}"
            )
        root = TestRoot(rname, Path(rsource))
        roots.append(root)
        rnames.add(root.name)

    # TODO: Consider allowing path to configuration file, pass here.
    configuration = conf.get_configuration()
    srvs = ManagedServices(state_dir, reuse_state=True, roots=roots, configuration=configuration)
    try:
        srvs.start_all()
        sub_auth.make_sub_user(srvs)
        srvs.wait_for_all()
    finally:
        srvs.stop_all()


if __name__ == "__main__":
    main()



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2025 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################
import math
from time import time

import caterva2 as cat2

# Choose the dataset to open
# dsname = 'examples/sa-1M.b2nd'
dsname = "examples/lung-jpeg2000_10x.b2nd"
# dsname = 'examples/uncompressed_lung-jpeg2000_10x.b2nd'

# Open a client to the Cat2Cloud server
client = cat2.Client("https://cat2.cloud/demo")
info = client.get_info(f"@public/{dsname}")
print(f"Client info: {info.keys()}")
typesize = info["schunk"]["cparams"]["typesize"]
# Open the public root
root = client.get("@public")
# Open the remote array
t0 = time()
remote_array = root[dsname]
size = math.prod(remote_array.shape) * typesize
print(f"Time for opening data (HTTP): {time() - t0:.3f}s - file size: {size / 2**10:.2f} KB")

# Download a slice of the remote array as a numpy array
t0 = time()
na = remote_array[5:9]
print(f"Time for reading data (getitem): {time() - t0:.3f}s - data size: {na.nbytes / 2**10:.2f} KB")

# Download a slice of the remote array
t0 = time()
a = remote_array.slice(slice(5, 9))
print(f"Time for reading data (slice): {time() - t0:.3f}s - data size: {a.schunk.nbytes / 2**10:.2f} KB")



def main(*args, **kwargs):
    """An alias for `caterva2.tests.__main__.main()`."""
    # Avoid `RuntimeWarning` about import order of ``__main__``
    # resulting in unpredictable behaviour.
    from caterva2.tests.__main__ import main as tests_main
    return tests_main(*args, **kwargs)



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################

"""
Add a user to the subscriber database.

Contrarily to `cat2cli adduser`, this script does not require a running Caterva2 subscriber.
"""

from caterva2 import utils
from caterva2.services import srv_utils


def main():
    conf = utils.get_conf("subscriber", allow_id=True)
    _stdir = "_caterva2/sub" + (f".{conf.id}" if conf.id else "")
    parser = utils.get_parser(statedir=conf.get(".statedir", _stdir), id=conf.id)
    parser.add_argument("username")
    parser.add_argument("password", nargs="?")
    parser.add_argument("--superuser", "-S", action="store_true", default=False)
    args = utils.run_parser(parser)

    statedir = args.statedir.resolve()
    user = srv_utils.add_user(args.username, args.password, args.superuser, state_dir=statedir)
    print("Password:", user.password)


if __name__ == "__main__":
    main()



import pathlib

# Requirements
import numpy
import PIL.Image
from fastapi import Depends, FastAPI, Request, responses
from fastapi.responses import HTMLResponse
from fastapi.templating import Jinja2Templates

# Project
from caterva2.services.sub import optional_user

from ...sub import get_container, open_b2, resize_image
from ...sub import templates as sub_templates
from ...subscriber import db

app = FastAPI()
BASE_DIR = pathlib.Path(__file__).resolve().parent
templates = Jinja2Templates(directory=BASE_DIR / "templates")

name = "tomography"  # Identifies the plugin
label = "Tomography"
contenttype = "tomography"


abspath_and_dataprep = None
urlbase = None


def init(absp_n_datap, urlbase_):
    global abspath_and_dataprep
    global urlbase
    abspath_and_dataprep = absp_n_datap
    urlbase = urlbase_


def url(path: str) -> str:
    return f"{urlbase}/{path}"


def guess(path: pathlib.Path, meta) -> bool:
    """Does dataset (given path and metadata) seem of this content type?"""
    if not hasattr(meta, "dtype"):
        return False  # not an array

    dtype = meta.dtype
    if dtype is None:
        return False

    # Structured dtype
    if isinstance(dtype, str) and dtype.startswith("["):
        dtype = eval(dtype)  # TODO Make it safer

    dtype = numpy.dtype(dtype)
    shape = tuple(meta.shape)
    if dtype.kind != "u":
        return False

    if len(shape) == 3:
        return True  # grayscale

    # RGB(A)
    return len(shape) == 4 and shape[-1] in (3, 4)


@app.get("/display/{path:path}", response_class=HTMLResponse)
def display(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    user: db.User = Depends(optional_user),
):
    abspath, _ = abspath_and_dataprep(path, user=user)
    arr = open_b2(abspath, path)

    base = url(f"plugins/{name}")
    context = {
        "href": f"{base}/display_one/{path}",
        "pages": arr.shape[0],
    }
    return templates.TemplateResponse(request, "display.html", context=context)


async def __get_image(path, user, i):
    # Alternatively, call abspath_and_dataprep with the corresponding slice to download data async
    array = await get_container(path, user)
    content = array[i]
    return PIL.Image.fromarray(content)


@app.get("/display_one/{path:path}", response_class=HTMLResponse)
async def display_one(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    # Query parameters
    i: int,
    user: db.User = Depends(optional_user),
):
    img = await __get_image(path, user, i)

    base = url(f"plugins/{name}")
    src = f"{base}/image/{path}?{i=}"

    width = 768  # Max size
    links = []
    if img.width > width:
        links.append(
            {
                "href": src,
                "label": f"{img.width} x {img.height} (original size)",
                "target": "blank_",
            }
        )
        src = f"{src}&{width=}"

    context = {"src": src, "links": links}
    return sub_templates.TemplateResponse(request, "display_image.html", context=context)


@app.get("/image/{path:path}")
async def image_file(
    request: Request,
    # Path parameters
    path: pathlib.Path,
    # Query parameters
    i: int,
    width: int | None = None,
    user: db.User = Depends(optional_user),
):
    img = await __get_image(path, user, i)
    img_file = resize_image(img, width)

    def iterfile():
        yield from img_file

    return responses.StreamingResponse(iterfile(), media_type="image/png")



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################


import json
import subprocess
import sys

import pytest

from .services import TEST_CATERVA2_ROOT


@pytest.fixture
def pub_host(services):
    return services.get_endpoint(f'publisher.{TEST_CATERVA2_ROOT}')


@pytest.fixture
def sub_urlbase(services):
    return services.get_urlbase('subscriber')


def cli(cargs, binary=False, sub_user=None) -> str or dict:
    cli_path = 'caterva2.clients.cli'
    args = [sys.executable, '-m' + str(cli_path)]
    if sub_user:
        args += ['--username', sub_user.username,
                 '--password', sub_user.password]
    args += cargs
    if not binary:
        args += ['--json']
    ret = subprocess.run(args, capture_output=True, text=True)
    assert ret.returncode == 0
    out = ret.stdout
    return out if binary else json.loads(out)


def test_roots(pub_host, sub_user):
    roots = cli(['roots'], sub_user=sub_user)
    assert roots[TEST_CATERVA2_ROOT]['name'] == TEST_CATERVA2_ROOT
    assert roots[TEST_CATERVA2_ROOT]['http'] == pub_host


def test_url(sub_urlbase, sub_user):
    out = cli(['url', f'{TEST_CATERVA2_ROOT}/ds-1d.b2nd'], sub_user=sub_user)
    assert out == f'{sub_urlbase}/api/fetch/{TEST_CATERVA2_ROOT}/ds-1d.b2nd'


def test_subscribe(sub_user):
    # Subscribe once
    out = cli(['subscribe', TEST_CATERVA2_ROOT], sub_user=sub_user)
    assert out == 'Ok'

    # Subscribe again, should be a noop
    out = cli(['subscribe', TEST_CATERVA2_ROOT], sub_user=sub_user)
    assert out == 'Ok'

    # Show
    a = cli(['show', f'{TEST_CATERVA2_ROOT}/ds-1d.b2nd'],
            binary=True, sub_user=sub_user)
    b = cli(['show', f'{TEST_CATERVA2_ROOT}/ds-1d.b2nd'],
            binary=True, sub_user=sub_user)
    assert a == b



###############################################################################
# Caterva2 - On demand access to remote Blosc2 data repositories
#
# Copyright (c) 2023 ironArray SLU <contact@ironarray.io>
# https://www.blosc.org
# License: GNU Affero General Public License v3.0
# See LICENSE.txt for details about copyright and rights to use.
###############################################################################
from time import time

import blosc2

# Small example of how to query a dataset on a subscriber
import httpx
import numpy as np

import caterva2 as cat2

# Use the demo server
URLBASE = "https://demo.caterva2.net/"
ROOT_NAME = "example"

user_auth = None
# Uncomment the following line and use your username and password
# if the subscriber requires authentication.
# user_auth = {'username': 'user@example.com', 'password': 'foobar11'}

client = cat2.Client(URLBASE, auth=user_auth)
# Get the list of available roots
roots = client.get_roots()
print(roots[ROOT_NAME])
# Get a handle to the root
example = client.get(ROOT_NAME)
# List the datasets in that root
print(example.file_list)
# Get a specific dataset
array = example["dir1/ds-2d.b2nd"]
print(array.name, array.path)
# Get some data out of the dataset
print(array[:2])

# There are different ways to get the data
# 1. Direct download
t0 = time()
urlpath = array.get_download_url()
data = httpx.get(urlpath, auth=user_auth)
t = time() - t0
# print(urlpath)
print(f"Time for downloading data (HTTP): {t:.3f}s - {len(data.content) / 2**10:.2f} KB")
mem_array = blosc2.ndarray_from_cframe(data.content)
print(mem_array.info)

# 2. Fetch the data from the subscriber
t0 = time()
mem_array2 = array[:]
t = time() - t0
print(f"Time for downloading data (API) : {t:.3f}s - {mem_array2.nbytes / 2**10:.2f} KB")

# Check that both methods return the same data
np.testing.assert_array_equal(mem_array2, mem_array[:])
